/*
 * Sample Driver for ak8854
 *
 * Copyright (C) 2011, Renesas Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/videodev2.h>
#include <linux/gpio.h>
#include <linux/firmware.h>

#include <media/soc_camera.h>
#include <media/soc_mediabus.h>
#include <media/v4l2-subdev.h>
#include <media/v4l2-chip-ident.h>

/* AK8854 supported geometry for NTSC */
#define AK8854_WIDTH		720
#define AK8854_HEIGHT		480
#define AK8854_OFFSET_LEFT	58
#define AK8854_OFFSET_TOP	30
#define AK8854_BOUNDS_WIDTH	858
#define AK8854_BOUNDS_HEIGHT	525


static const u8 i2c_init_data_NTSC[][2] = {
	{0x07, 0x40},
};


struct ak8854_datafmt {
	enum v4l2_mbus_pixelcode	code;
	enum v4l2_colorspace		colorspace;
};

struct ak8854 {
	struct v4l2_subdev		subdev;
	const struct ak8854_datafmt	*fmt;
	enum v4l2_field			field;
};

static const struct ak8854_datafmt ak8854_colour_fmts[] = {
	{V4L2_MBUS_FMT_UYVY8_2X8, V4L2_COLORSPACE_SMPTE170M},
};

static struct ak8854 *
to_ak8854(const struct i2c_client *client)
{
	return container_of(i2c_get_clientdata(client), struct ak8854, subdev);
}

/* Find a data format by a pixel code in an array */
static const struct ak8854_datafmt *
ak8854_find_datafmt(enum v4l2_mbus_pixelcode code)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(ak8854_colour_fmts); i++)
		if (ak8854_colour_fmts[i].code == code)
			return ak8854_colour_fmts + i;

	return NULL;
}

static int
ak8854_try_fmt(struct v4l2_subdev *sd,
	       struct v4l2_mbus_framefmt *mf)
{
	const struct ak8854_datafmt *fmt = ak8854_find_datafmt(mf->code);

	dev_dbg(sd->v4l2_dev->dev, "%s(%u), %ux%u, %u\n", __func__, mf->code,
		mf->width, mf->height, mf->field);

	if (!fmt) {
		mf->code	= ak8854_colour_fmts[0].code;
		mf->colorspace	= ak8854_colour_fmts[0].colorspace;
	}

	mf->width	= AK8854_WIDTH;
	mf->height	= AK8854_HEIGHT;
/*	mf->field	= V4L2_FIELD_NONE;*/

	if (mf->field == V4L2_FIELD_TOP ||
	    mf->field == V4L2_FIELD_BOTTOM ||
	    mf->field == V4L2_FIELD_ALTERNATE)
		mf->height /= 2;

	return 0;
}

static int
ak8854_init(struct v4l2_subdev *sd, const u8 (*data)[2], int len)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	int i;

	for (i = 0; i < len; i++) {
		if (i2c_master_send(client, data[i], 2) != 2) {
			dev_err(sd->v4l2_dev->dev, "i2c write error\n");
			return -EIO;
		}
	}

	return 0;
}

static int
ak8854_s_fmt(struct v4l2_subdev *sd,
	     struct v4l2_mbus_framefmt *mf)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct ak8854 *priv = to_ak8854(client);
	int ret;

	dev_dbg(sd->v4l2_dev->dev, "%s(%u)\n", __func__, mf->code);

	/* MIPI CSI could have changed the format, double-check */
	if (!ak8854_find_datafmt(mf->code)) {
		dev_err(sd->v4l2_dev->dev, "%s -EINVAL\n", __func__);
		return -EINVAL;
	}

	ret = ak8854_try_fmt(sd, mf);
	if (ret)
		return ret;

	priv->fmt = ak8854_find_datafmt(mf->code);
	priv->field = mf->field;

	ak8854_init(sd, i2c_init_data_NTSC, ARRAY_SIZE(i2c_init_data_NTSC));

	return 0;
}

static int
ak8854_g_fmt(struct v4l2_subdev *sd,
	     struct v4l2_mbus_framefmt *mf)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct ak8854 *priv = to_ak8854(client);

	const struct ak8854_datafmt *fmt = priv->fmt;

	dev_dbg(sd->v4l2_dev->dev, "%s\n", __func__);

	mf->code	= fmt->code;
	mf->colorspace	= fmt->colorspace;
	mf->width	= AK8854_WIDTH;
	mf->height	= AK8854_HEIGHT;
	mf->field	= priv->field;

	if (priv->field == V4L2_FIELD_TOP ||
	    priv->field == V4L2_FIELD_BOTTOM ||
	    priv->field == V4L2_FIELD_ALTERNATE)
		mf->height /= 2;

	return 0;
}

static int
ak8854_g_crop(struct v4l2_subdev *sd, struct v4l2_crop *a)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct ak8854 *priv = to_ak8854(client);
	struct v4l2_rect *rect = &a->c;

	dev_dbg(sd->v4l2_dev->dev, "%s\n", __func__);

	a->type		= V4L2_BUF_TYPE_VIDEO_CAPTURE;
	rect->left	= 0;
	rect->top	= 0;
	rect->width	= AK8854_WIDTH;
	rect->height	= AK8854_HEIGHT;

	if (priv->field == V4L2_FIELD_TOP ||
	    priv->field == V4L2_FIELD_BOTTOM ||
	    priv->field == V4L2_FIELD_ALTERNATE)
		rect->height /= 2;

	return 0;
}

static int
ak8854_cropcap(struct v4l2_subdev *sd, struct v4l2_cropcap *a)
{
	dev_dbg(sd->v4l2_dev->dev, "%s\n", __func__);

	if (a->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	a->bounds.left			= 0;
	a->bounds.top			= 0;
	a->bounds.width			= AK8854_BOUNDS_WIDTH;
	a->bounds.height		= AK8854_BOUNDS_HEIGHT;
	a->defrect.left			= AK8854_OFFSET_LEFT;
	a->defrect.top			= AK8854_OFFSET_TOP;
	a->defrect.width		= AK8854_WIDTH;
	a->defrect.height		= AK8854_HEIGHT;
	a->type				= V4L2_BUF_TYPE_VIDEO_CAPTURE;
	a->pixelaspect.numerator	= 1;
	a->pixelaspect.denominator	= 1;

	return 0;
}

static int
ak8854_enum_fmt(struct v4l2_subdev *sd, unsigned int index,
		enum v4l2_mbus_pixelcode *code)
{
	if ((unsigned int)index >= ARRAY_SIZE(ak8854_colour_fmts))
		return -EINVAL;

	*code = ak8854_colour_fmts[index].code;
	return 0;
}

static int
ak8854_s_stream(struct v4l2_subdev *sd, int enable)
{
	struct i2c_client *client = sd->priv;
	struct soc_camera_device *icd = client->dev.platform_data;
	struct soc_camera_link *icl = to_soc_camera_link(icd);
	int ret;

	if (!icl->power)
		return -1;
	ret = icl->power(icd->pdev, enable ? 2 : 1);

	return ret;
}

static int
ak8854_g_chip_ident(struct v4l2_subdev *sd,
		    struct v4l2_dbg_chip_ident *id)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);

	if (id->match.type != V4L2_CHIP_MATCH_I2C_ADDR)
		return -EINVAL;

	if (id->match.addr != client->addr)
		return -ENODEV;

	id->ident	= V4L2_IDENT_UNKNOWN/*V4L2_IDENT_AK8854*/;
	id->revision	= 0;

	return 0;
}

static struct v4l2_subdev_video_ops ak8854_subdev_video_ops = {
	.s_stream	= ak8854_s_stream,
	.s_mbus_fmt	= ak8854_s_fmt,
	.g_mbus_fmt	= ak8854_g_fmt,
	.try_mbus_fmt	= ak8854_try_fmt,
	.enum_mbus_fmt	= ak8854_enum_fmt,
	.g_crop		= ak8854_g_crop,
	.cropcap	= ak8854_cropcap,
};

static struct v4l2_subdev_core_ops ak8854_subdev_core_ops = {
	.g_chip_ident	= ak8854_g_chip_ident,
};

static struct v4l2_subdev_ops ak8854_subdev_ops = {
	.core	= &ak8854_subdev_core_ops,
	.video	= &ak8854_subdev_video_ops,
};

static unsigned long
ak8854_query_bus_param(struct soc_camera_device *icd)
{
	struct soc_camera_link *icl = to_soc_camera_link(icd);
	unsigned long flags = SOCAM_PCLK_SAMPLE_RISING | SOCAM_MASTER |
		SOCAM_VSYNC_ACTIVE_HIGH | SOCAM_HSYNC_ACTIVE_HIGH |
		SOCAM_DATA_ACTIVE_HIGH;

	flags |= SOCAM_DATAWIDTH_8;

	return soc_camera_apply_sensor_flags(icl, flags);
}

static int
ak8854_set_bus_param(struct soc_camera_device *icd,
		     unsigned long flags)
{
	return 0;
}

static struct soc_camera_ops ak8854_ops = {
	.query_bus_param	= ak8854_query_bus_param,
	.set_bus_param		= ak8854_set_bus_param,
};

static int ak8854_probe(struct i2c_client *client,
			const struct i2c_device_id *did)
{
	struct ak8854 *priv;
	struct soc_camera_device *icd = client->dev.platform_data;
	struct i2c_adapter *adapter = to_i2c_adapter(client->dev.parent);
	struct soc_camera_link *icl;
	int ret = 0;

	if (!icd) {
		dev_err(&client->dev, "AK8854: missing soc-camera data!\n");
		return -EINVAL;
	}

	icl = to_soc_camera_link(icd);
	if (!icl) {
		dev_err(&client->dev, "AK8854: missing platform data!\n");
		return -EINVAL;
	}

	if (!i2c_check_functionality(adapter, I2C_FUNC_I2C)) {
		dev_warn(&adapter->dev,
			 "I2C-Adapter doesn't support I2C_FUNC_I2C\n");
		return -EIO;
	}

	priv = kzalloc(sizeof(struct ak8854), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	v4l2_i2c_subdev_init(&priv->subdev, client, &ak8854_subdev_ops);

	icd->ops	= &ak8854_ops;
	priv->fmt	= &ak8854_colour_fmts[0];


	if (icl->free_bus)
		icl->free_bus(icl);

	return ret;
}

static int ak8854_remove(struct i2c_client *client)
{
	struct ak8854 *priv = to_ak8854(client);
	struct soc_camera_device *icd = client->dev.platform_data;
	struct soc_camera_link *icl = to_soc_camera_link(icd);

	icd->ops = NULL;
	if (icl->free_bus)
		icl->free_bus(icl);
	kfree(priv);

	return 0;
}

static const struct i2c_device_id ak8854_id[] = {
	{ "ak8854", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, ak8854_id);

static struct i2c_driver ak8854_i2c_driver = {
	.driver = {
		.name = "ak8854",
	},
	.probe		= ak8854_probe,
	.remove		= ak8854_remove,
	.id_table	= ak8854_id,
};

static int __init ak8854_mod_init(void)
{
	return i2c_add_driver(&ak8854_i2c_driver);
}

static void __exit ak8854_mod_exit(void)
{
	i2c_del_driver(&ak8854_i2c_driver);
}

module_init(ak8854_mod_init);
module_exit(ak8854_mod_exit);

MODULE_DESCRIPTION("AK8854 sample driver");
MODULE_AUTHOR("Renesas Electronics Corporation");
MODULE_LICENSE("GPL v2");
