/*
 *  EHCI HCD (Host Controller Driver) for USB.
 *
 *  Copyright (C) 2010-2011 Renesas Electronics Corporation
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

#include <linux/clk.h>
#include <linux/device.h>
#include <linux/signal.h>
#include <linux/platform_device.h>
#include <linux/miscdevice.h>
#include <linux/workqueue.h>
#include <linux/usb/rmobile.h>
#include <linux/pm_runtime.h>

struct rmobile_ehci {
	/* must be 1st member here for hcd_to_ehci() to work */
	struct ehci_hcd ehci;

	u32		phys_base;
	void __iomem	*base;

	int		vbus_flag;
	u32		relinquish_count;
	u32		handed_over;

	/* platform data */
	struct rmobile_usb_platdata *pdata;

	/* backup area */
	struct rmobile_ehci_reg ehci_reg;
};

#define to_rmobile_ehci(hcd)	(struct rmobile_ehci *)hcd_to_ehci(hcd)

/*-------------------------------------------------------------------------*/

static void rmobile_vbus_control(struct rmobile_ehci *ehci, int on)
{
	if (ehci->pdata->set_vbus)
		ehci->pdata->set_vbus(on);
	ehci->vbus_flag = on ? 0 : 1;
}

static int rmobile_over_current_check(struct rmobile_ehci *ehci)
{
	if (ehci->pdata->check_oci()) {
		rmobile_vbus_control(ehci, 0);
		err("***** USB Over Current !!\n");
		return 1;
	}
	return 0;
}

/*-------------------------------------------------------------------------*/

static void rmobile_hc_start(struct usb_hcd *hcd)
{
	u32 data, tmp;
	struct clk *clk;
	struct rmobile_ehci *ehci = to_rmobile_ehci(hcd);

	if (ehci->pdata->common_clock)
		ehci->pdata->common_clock(1);
	else {
		/* Start Clock */
		clk = clk_get(NULL, "usbh");
		if (!IS_ERR(clk)) {
			clk_enable(clk);
			clk_put(clk);
		}
		clk = clk_get(NULL, "usbp");
		if (!IS_ERR(clk)) {
			clk_enable(clk);
			clk_put(clk);
		}

		/* Clock & Reset & Direct Power Down */
		data = __raw_readl(ehci->base + USBCTR);
		data &= ~(DIRPD | PCICLK_MASK | USBH_RST);
		__raw_writel(data, ehci->base + USBCTR);
		__raw_writel(data & ~PLL_RST, ehci->base + USBCTR);
	}

	data = SMODE_READY_CTR | MMODE_HBUSREQ |
			MMODE_WR_INCR | MMODE_BYTE_BURST | MMODE_HTRANS;

	tmp = __raw_readl(ehci->base + AHB_BUS_CTR);
	if (data == tmp)
		return;

	/****** AHB-PCI Bridge Communication Registers ******/
	__raw_writel(data, ehci->base + AHB_BUS_CTR);
	__raw_writel((PHYS_OFFSET & 0xf0000000) | PREFETCH,
			ehci->base + PCIAHB_WIN1_CTR);
	__raw_writel(0xf0000000 | PREFETCH, ehci->base + PCIAHB_WIN2_CTR);
	__raw_writel(ehci->phys_base | PCIWIN2_PCICMD,
			ehci->base + AHBPCI_WIN2_CTR);
	data = __raw_readl(ehci->base + PCI_ARBITER_CTR);
	__raw_writel(data | PCIBP_MODE | PCIREQ1 | PCIREQ0,
			ehci->base + PCI_ARBITER_CTR);

	/****** PCI Configuration Registers for AHBPCI ******/
	__raw_writel(PCIWIN1_PCICMD | AHB_CFG_AHBPCI,
			ehci->base + AHBPCI_WIN1_CTR);
	__raw_writel(ehci->phys_base + AHBPCI_OFFSET, ehci->base + BASEAD);
	__raw_writel(PHYS_OFFSET & 0xf0000000, ehci->base + WIN1_BASEAD);
	__raw_writel(0xf0000000, ehci->base + WIN2_BASEAD);
	__raw_writel(SERREN | PERREN | MASTEREN | MEMEN,
			ehci->base + CMND_STS);

	/****** PCI Configuration Registers for ehci/EHCI ******/
	data = PCIWIN1_PCICMD | AHB_CFG_HOST;
	__raw_writel(data, ehci->base + AHBPCI_WIN1_CTR);
	__raw_writel(ehci->phys_base + OHCI_OFFSET, ehci->base + OHCI_BASEAD);
	__raw_writel(ehci->phys_base + EHCI_OFFSET, ehci->base + EHCI_BASEAD);
	data = SERREN | PERREN | MASTEREN | MEMEN;
	__raw_writel(data, ehci->base + OHCI_CMND_STS);
	__raw_writel(data, ehci->base + EHCI_CMND_STS);

	/* PCI_INT_ENABLE */
	data = __raw_readl(ehci->base + PCI_INT_ENABLE);
	data |= USBH_PMEEN | USBH_INTBEN | USBH_INTAEN;
	__raw_writel(data | USBH_PMEEN | USBH_INTBEN | USBH_INTAEN,
			ehci->base + PCI_INT_ENABLE);
}

static void rmobile_hc_stop(struct rmobile_ehci *ehci)
{
	u32 data;
	struct clk *clk;

	rmobile_vbus_control(ehci, 0);

	if (ehci->pdata->common_clock)
		ehci->pdata->common_clock(0);
	else {
		/* Clock & Reset & Direct Power Down */
		data = __raw_readl(ehci->base + USBCTR);
		data |= DIRPD | PLL_RST | PCICLK_MASK | USBH_RST;
		__raw_writel(data, ehci->base + USBCTR);

		/* Stop Clock */
		clk = clk_get(NULL, "usbh");
		if (!IS_ERR(clk)) {
			clk_disable(clk);
			clk_put(clk);
		}
		clk = clk_get(NULL, "usbp");
		if (!IS_ERR(clk)) {
			clk_disable(clk);
			clk_put(clk);
		}
	}
}

/*-------------------------------------------------------------------------*/

static void rmobile_relinquish_port(struct usb_hcd *hcd, int portnum)
{
	struct rmobile_ehci *ehci = to_rmobile_ehci(hcd);

	if ((ehci->handed_over == 0) && (ehci->relinquish_count == 0)) {
		pr_err(" --- Reset Host Controller\n");
		ehci->relinquish_count++;
		rmobile_vbus_control(ehci, 0);
		if (ehci->pdata->oci_irq)
			disable_irq(ehci->pdata->oci_irq);
		/* rmobile_hc_fatal_recovery(hcd); */
		return;
	}

	ehci->handed_over = 0;
	ehci->relinquish_count = 0;
	ehci_relinquish_port(hcd, portnum);
}


static int rmobile_port_handed_over(struct usb_hcd *hcd, int portnum)
{
	struct rmobile_ehci *ehci = to_rmobile_ehci(hcd);

	ehci->handed_over = 1;

	return ehci_port_handed_over(hcd, portnum);
}

/*-------------------------------------------------------------------------*/

static int rmobile_hub_control(
	struct usb_hcd	*hcd,
	u16		typeReq,
	u16		wValue,
	u16		wIndex,
	char		*buf,
	u16		wLength
)
{
	int retval;
	struct rmobile_ehci *ehci = to_rmobile_ehci(hcd);

	retval = ehci_hub_control(hcd, typeReq, wValue, wIndex, buf, wLength);

	/* VBUS ON/OFF control */
	if (retval == 0) {
		if (wValue == USB_PORT_FEAT_POWER) {
			if (typeReq == SetPortFeature) {
				rmobile_vbus_control(ehci, 1);
				if (ehci->pdata->oci_irq) {
					mdelay(OVER_CURRENT_TIME);
					if (!rmobile_over_current_check(ehci))
						enable_irq(ehci->pdata->oci_irq);
				}
			} else if (typeReq == ClearPortFeature) {
				if (ehci->vbus_flag) {
					if (ehci->pdata->oci_irq)
						disable_irq(ehci->pdata->oci_irq);
					rmobile_vbus_control(ehci, 0);
				} else
					ehci->vbus_flag = 1;
			}
		}
	}

	return retval;
}

static irqreturn_t rmobile_over_current_irq(int irq, void *_hcd)
{
	struct rmobile_ehci *ehci = to_rmobile_ehci(_hcd);
	rmobile_over_current_check(ehci);

	return IRQ_HANDLED;
}

/*-------------------------------------------------------------------------*/

static void rmobile_ehci_save_register(struct usb_hcd *hcd)
{
	struct rmobile_ehci_reg *p_reg;
	struct ehci_hcd *ehci = hcd_to_ehci(hcd);
	struct rmobile_ehci *rehci = to_rmobile_ehci(hcd);

	p_reg = &rehci->ehci_reg;
	if (p_reg->save_flag == 0)
		return;

	p_reg->save_flag = 1;

	p_reg->USBINTR = ehci_readl(ehci, &ehci->regs->intr_enable);
	p_reg->USBCMD = ehci_readl(ehci, &ehci->regs->command);
	p_reg->USBSTS = ehci_readl(ehci, &ehci->regs->status);
	p_reg->FRINDEX = ehci_readl(ehci, &ehci->regs->frame_index);
	p_reg->PERIODICLISTBASE = ehci_readl(ehci, &ehci->regs->frame_list);
	p_reg->ASYNCLISTADDR = ehci_readl(ehci, &ehci->regs->async_next);
	p_reg->CONFIGFLAG = ehci_readl(ehci, &ehci->regs->configured_flag);
	p_reg->PORTSC = ehci_readl(ehci, &ehci->regs->port_status[0]);
}

static void rmobile_ehci_load_register(struct usb_hcd *hcd)
{
	struct rmobile_ehci_reg *p_reg;
	struct ehci_hcd *ehci = hcd_to_ehci(hcd);
	struct rmobile_ehci *rehci = to_rmobile_ehci(hcd);

	p_reg = &rehci->ehci_reg;
	if (p_reg->save_flag == 0)
		return;

	ehci_writel(ehci, p_reg->CONFIGFLAG, &ehci->regs->configured_flag);
	ehci_writel(ehci, p_reg->PORTSC, &ehci->regs->port_status[0]);
	ehci_writel(ehci, p_reg->ASYNCLISTADDR, &ehci->regs->async_next);
	ehci_writel(ehci, p_reg->PERIODICLISTBASE, &ehci->regs->frame_list);
	ehci_writel(ehci, p_reg->FRINDEX, &ehci->regs->frame_index);
	ehci_writel(ehci, p_reg->USBSTS, &ehci->regs->status);
	ehci_writel(ehci, p_reg->USBCMD, &ehci->regs->command);
	ehci_writel(ehci, p_reg->USBINTR, &ehci->regs->intr_enable);

	p_reg->save_flag = 0;
}

/*-------------------------------------------------------------------------*/
static ssize_t _portsc_show(struct device *dev,
			    struct device_attribute *attr, char *buf)
{
	struct usb_hcd *hcd = dev_get_drvdata(dev);
	struct ehci_hcd *ehci = hcd_to_ehci(hcd);
	return sprintf(buf, "0x%08x\n", ehci_readl(ehci, &ehci->regs->port_status[0]));
}
static ssize_t _portsc_store(struct device *dev,
			     struct device_attribute *attr,
			     const char *buf, size_t size)
{
	u32 val = simple_strtoul(buf, NULL, 0);
	struct usb_hcd *hcd = dev_get_drvdata(dev);
	struct ehci_hcd *ehci = hcd_to_ehci(hcd);
	ehci_writel(ehci, val, &ehci->regs->port_status[0]);
	return size;
}

static DEVICE_ATTR(portsc, 0644, _portsc_show, _portsc_store);

extern int usb_get_device_descriptor(struct usb_device *dev,
				     unsigned int size);
static ssize_t _single_step_get_dev_desc_store(struct device *dev,
					       struct device_attribute *attr,
					       const char *buf, size_t size)
{
	struct usb_hcd *hcd = dev_get_drvdata(dev);
	struct usb_bus *ubus = hcd_to_bus(hcd);
	struct usb_device *udev;

	if (!ubus) {
		printk(KERN_ERR "ERROR: usb_bus not found\n");
		return size;
	}
	udev = ubus->root_hub->children[0];
	if (!udev) {
		printk(KERN_ERR "ERROR: usb_device not found\n");
		return size;
	}
	usb_get_device_descriptor(udev, sizeof(struct usb_device_descriptor));

	return size;
}

static DEVICE_ATTR(single_get, 0200, NULL, _single_step_get_dev_desc_store);

static ssize_t _single_step_set_feature_store(struct device *dev,
					      struct device_attribute *attr,
					      const char *buf, size_t size)
{
	struct usb_hcd *hcd = dev_get_drvdata(dev);
	struct usb_bus *ubus = hcd_to_bus(hcd);
	struct usb_device *udev;

	if (!ubus) {
		printk(KERN_ERR "ERROR: usb_bus not found\n");
		return size;
	}
	udev = ubus->root_hub->children[0];
	if (!udev) {
		printk(KERN_ERR "ERROR: usb_device not found\n");
		return size;
	}
	usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
			USB_REQ_SET_FEATURE, 0, 0,
			0, NULL, 0, HZ * USB_CTRL_SET_TIMEOUT);

	return size;
}

static DEVICE_ATTR(single_set, 0200, NULL, _single_step_set_feature_store);

static int rmobile_hcd_probe(
	const struct hc_driver *driver,
	struct platform_device *pdev
)
{
	int retval, irq;
	struct usb_hcd *hcd;
	struct ehci_hcd *ehci;
	struct rmobile_ehci *rehci;
	struct rmobile_usb_platdata *pdata;
	struct resource *r;

	pdata = pdev->dev.platform_data;
	if (!pdata)
		return -ENODEV;

	/* OHCI irq */
	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		pr_err("no resource of IORESOURCE_IRQ");
		return -ENXIO;
	}

	hcd = usb_create_hcd(driver, &pdev->dev, "rmobile_ehci");
	if (!hcd)
		return -ENOMEM;

	r = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!r) {
		pr_err("no resource of IORESOURCE_MEM");
		retval = -ENXIO;
		goto err1;
	}

	hcd->rsrc_start = r->start + EHCI_OFFSET;
	hcd->rsrc_len = EHCI_SIZE;
	if (!request_mem_region(hcd->rsrc_start, hcd->rsrc_len, hcd_name)) {
		pr_debug("request_mem_region failed");
		retval = -EBUSY;
		goto err1;
	}

	hcd->regs = ioremap(r->start, resource_size(r));
	if (!hcd->regs) {
		pr_debug("ioremap failed");
		retval = -ENOMEM;
		goto err2;
	}

	rehci = (struct rmobile_ehci *)hcd_to_ehci(hcd);
	rehci->phys_base = r->start;
	rehci->base = hcd->regs;
	hcd->regs += EHCI_OFFSET;
	rehci->pdata = pdata;
	if (pdata->oci_irq) {
		if (!pdata->check_oci) {
			retval = -ENODEV;
			goto err3;
		} else {
			retval = request_irq(pdata->oci_irq,
				rmobile_over_current_irq,
				IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING |
					IRQF_SHARED, "Over Current", hcd);
			if (!retval)
				goto err3;
			disable_irq(pdata->oci_irq);
		}
	}

	pm_suspend_ignore_children(&pdev->dev, true);
	pm_runtime_enable(&pdev->dev);
	pm_runtime_get_noresume(&pdev->dev);

	rmobile_hc_start(hcd);

	ehci = hcd_to_ehci(hcd);
	ehci->caps = hcd->regs;
	ehci->regs = hcd->regs + HC_LENGTH(readl(&ehci->caps->hc_capbase));

	/* cache this readonly data; minimize chip reads */
	ehci->hcs_params = readl(&ehci->caps->hcs_params);
	ehci->sbrn = 0x20;
	hcd->has_tt = 0;

	if (device_create_file(&pdev->dev, &dev_attr_portsc))
		dev_err(&pdev->dev, "failed create PORTSC node\n");
	if (device_create_file(&pdev->dev, &dev_attr_single_get))
		dev_err(&pdev->dev, "failed create SINGLE_GET node\n");
	if (device_create_file(&pdev->dev, &dev_attr_single_set))
		dev_err(&pdev->dev, "failed create SINGLE_SET node\n");

	/* ehci_hcd_init(hcd_to_ehci(hcd)); */
	retval = usb_add_hcd(hcd, irq, IRQF_SHARED);
	if (retval == 0)
		return retval;

	rmobile_hc_stop(rehci);

	pm_runtime_disable(&pdev->dev);

	if (pdata->oci_irq)
		free_irq(pdata->oci_irq, hcd);

err3:
	iounmap(hcd->regs);
err2:
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
err1:
	usb_put_hcd(hcd);

	return retval;
}

static void rmobile_hcd_remove(
	struct usb_hcd *hcd,
	struct platform_device *pdev
)
{
	struct rmobile_ehci *ehci = to_rmobile_ehci(hcd);

	usb_remove_hcd(hcd);
	rmobile_hc_stop(ehci);
	if (ehci->pdata->oci_irq)
		free_irq(ehci->pdata->oci_irq, hcd);
	iounmap(ehci->base);
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
	usb_put_hcd(hcd);
	pm_runtime_disable(&pdev->dev);
}

/*-------------------------------------------------------------------------*/

static const struct hc_driver ehci_rmobile_hc_driver = {
	.description		= hcd_name,
	.product_desc		= "R-Mobile EHCI",
	.hcd_priv_size		= sizeof(struct rmobile_ehci),

	/*
	 * generic hardware linkage
	 */
	.irq			= ehci_irq,
	.flags			= HCD_MEMORY | HCD_USB2,

	/*
	 * basic lifecycle operations
	 */
	.reset			= ehci_init,
	.start			= ehci_run,
	.stop			= ehci_stop,
	.shutdown		= ehci_shutdown,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue		= ehci_urb_enqueue,
	.urb_dequeue		= ehci_urb_dequeue,
	.endpoint_disable	= ehci_endpoint_disable,
	.endpoint_reset		= ehci_endpoint_reset,

	/*
	 * scheduling support
	 */
	.get_frame_number	= ehci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data	= ehci_hub_status_data,
	.hub_control		= rmobile_hub_control,
	.bus_suspend		= ehci_bus_suspend,
	.bus_resume		= ehci_bus_resume,
	.relinquish_port	= rmobile_relinquish_port,
	.port_handed_over	= rmobile_port_handed_over,

	.clear_tt_buffer_complete = ehci_clear_tt_buffer_complete,
};

/*-------------------------------------------------------------------------*/

static int ehci_hcd_drv_probe(struct platform_device *pdev)
{
	if (usb_disabled())
		return -ENODEV;

	return rmobile_hcd_probe(&ehci_rmobile_hc_driver, pdev);
}

static int ehci_hcd_drv_remove(struct platform_device *pdev)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);

	rmobile_hcd_remove(hcd, pdev);

	return 0;
}

#ifdef CONFIG_PM
static int ehci_hcd_drv_suspend(struct device *dev)
{
	struct usb_hcd *hcd = dev_get_drvdata(dev);
	struct rmobile_ehci *ehci = to_rmobile_ehci(hcd);

	if (ehci->pdata->oci_irq)
		disable_irq(ehci->pdata->oci_irq);

	rmobile_vbus_control(ehci, 0);
	rmobile_ehci_save_register(hcd);
	rmobile_hc_stop(ehci);

	pm_runtime_put(dev);

	return 0;
}

static int ehci_hcd_drv_resume(struct device *dev)
{
	struct usb_hcd *hcd = dev_get_drvdata(dev);
	struct rmobile_ehci *ehci = to_rmobile_ehci(hcd);

	pm_runtime_get_sync(dev);

	rmobile_hc_start(hcd);
	rmobile_ehci_load_register(hcd);
	rmobile_vbus_control(ehci, 1);

	return 0;
}
static const struct dev_pm_ops ehci_hcd_pm_ops = {
	.suspend	= ehci_hcd_drv_suspend,
	.resume		= ehci_hcd_drv_resume,
};
#endif	/* CONFIG_PM */

static struct platform_driver ehci_hcd_rmobile_driver = {
	.probe = ehci_hcd_drv_probe,
	.remove = ehci_hcd_drv_remove,
	.shutdown = usb_hcd_platform_shutdown,
	.driver = {
		.name = "rmobile-ehci-driver",
		.owner	= THIS_MODULE,
#ifdef CONFIG_PM
		.pm	= &ehci_hcd_pm_ops,
#endif
	}
};

MODULE_ALIAS("platform:rmobile-ehci-driver");
