/*
 * R-Mobile A1 LCDC Framebuffer
 *
 * Copyright (c) 2011 Atmark Techno, Inc.
 *
 * Based on linux/drivers/video/sh_mobile_lcdcfb.c
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/fb.h>
#include <linux/clk.h>
#include <linux/pm_runtime.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>
#include <linux/vmalloc.h>
#include <linux/ioctl.h>
#include <linux/slab.h>
#include <linux/lcm.h>
#include <linux/gcd.h>
#include <video/sh_mobile_lcdc.h>
#include <asm/atomic.h>

#ifdef DEBUG
#define trace() { if (1) printk("*** %s()\n", __func__); 0; }
#define d_dbg(fmt, args...) { if (1) printk(fmt, args); 0; }
#else
#define trace() { if (0) printk("*** %s()\n", __func__); 0; }
#define d_dbg(fmt, args...) { if (0) printk(fmt, args); 0; }
#endif

#define PALETTE_NR 16
#define SETB_OFFSET	0x1000

/* registers */
#define LDDCKPAT1R	0x400
#define LDDCKPAT2R	0x404
#define LDDCKR		0x410
#define LDDCKSTPR	0x414
#define LDMT1R		0x418
#define LDMT2R		0x41c
#define LDMT3R		0x420
#define LDDFR		0x424
#define LDSA1R		0x430
#define LDSA2R		0x434
#define LDMLSR		0x438
#define LDHCNR		0x448
#define LDHSYNR		0x44c
#define LDVLNR		0x450
#define LDVSYNR		0x454
#define LDHPDR		0x458
#define LDVPDR		0x45c
#define LDPMR		0x460
#define LDINTR		0x468
#define LDSR		0x46c
#define LDCNT1R		0x470
#define LDCNT2R		0x474
#define LDRCNTR		0x478
#define LDDDSR		0x47c

#define START_LCDC	0x00000001
#define LCDC_RESET	0x00000100

#define priv_to_pdata(priv) (priv)->pdev->dev.platform_data
#define is_yuv_mode(nonstd) (nonstd == SH_FIXUP_YUV420) ? 1 : 0

#define s_writel(val, addr) \
	writel(val, addr)
#define s_writel_seta(val, addr) \
	s_writel(val, addr)
#define s_writel_setb(val, addr) \
	s_writel(val, addr + SETB_OFFSET)
#define s_writel_both(val, addr) \
({	s_writel_seta(val, addr); s_writel_setb(val, addr); 0; })

struct sh_mobile_lcdc_priv {
	struct fb_info *info;
	struct platform_device *pdev;

	void __iomem *base;
	int irq;

	atomic_t hw_usecnt;

	int dot_clk_source;
	struct clk *dot_clk;

	u32 pseudo_palette[PALETTE_NR];
};

static char *videomode;

static struct sh_videomode_ext *sh_find_videomode_ext(struct fb_info *info)
{
	struct sh_mobile_lcdc_priv *priv = info->par;
	struct sh_mobile_lcdc_info *pdata = priv_to_pdata(priv);
	struct fb_videomode *modedb = pdata->ch[0].modedb;
	const struct fb_videomode *mode;
	struct sh_videomode_ext *modedb_ext = pdata->ch[0].modedb_ext;
	unsigned int modedb_size = pdata->ch[0].modedb_size;
	unsigned int i;

	trace();

	if (!modedb || !modedb_ext)
		return NULL;

	mode = fb_match_mode(&info->var, &info->modelist);
	if (mode == NULL)
		return NULL;

	for (i = 0; i < modedb_size; i++)
		if ((strcmp(mode->name, modedb[i].name) == 0))
			return &modedb_ext[i];

	return NULL;
}

static void sh_mobile_lcdc_clk_on(struct sh_mobile_lcdc_priv *priv)
{
	struct platform_device *pdev = priv->pdev;
	struct clk *clk;

	trace();

	if (atomic_inc_and_test(&priv->hw_usecnt)) {
		pm_runtime_get_sync(&priv->pdev->dev);
		clk = clk_get(&pdev->dev, NULL);
		clk_enable(clk);
		clk_put(clk);
		if (priv->dot_clk)
			clk_enable(priv->dot_clk);
	}
}

static void sh_mobile_lcdc_clk_off(struct sh_mobile_lcdc_priv *priv)
{
	struct platform_device *pdev = priv->pdev;
	struct clk *clk;

	trace();

	if (atomic_sub_return(1, &priv->hw_usecnt) == -1) {
		clk = clk_get(&pdev->dev, NULL);
		clk_disable(clk);
		clk_put(clk);
		if (priv->dot_clk)
			clk_disable(priv->dot_clk);
		pm_runtime_put(&priv->pdev->dev);
	}
}

static void sh_mobile_lcdc_start_stop(struct sh_mobile_lcdc_priv *priv,
				      int start)
{
	unsigned long tmp = readl(priv->base + LDCNT2R);

	trace();

	/* start or stop the lcdc */
	if (start)
		writel(tmp | START_LCDC, priv->base + LDCNT2R);
	else
		writel(tmp & ~START_LCDC, priv->base + LDCNT2R);

	if (readl(priv->base + LDCNT2R) & 0x2) {
		while (1) {
			tmp = readl(priv->base + LDPMR) & 0x3;
			if (start && tmp == 0x3)
				break;
			if (!start && tmp == 0)
				break;
			cpu_relax();
		}
	}

	if (!start)
		writel(1, priv->base + LDDCKSTPR); /* stop dotclock */
	else
		writel(0, priv->base + LDDCKSTPR); /* start dotclock */
}

static int sh_mobile_lcdc_setup_clocks(struct sh_mobile_lcdc_priv *priv,
				       int clock_source)

{
	struct platform_device *pdev = priv->pdev;
	int icksel;
	char *clkstr;
	u32 val;

	trace();

	switch (clock_source) {
	case LCDC_CLK_BUS: clkstr = "m3_clk"; icksel = 0; break;
	case LCDC_CLK_HDMI: clkstr = "hdmi_clk"; icksel = 1; break;
	case LCDC_CLK_DV: clkstr = "dv_clk"; icksel = 3; break;
	default:
		return -EINVAL;
	}

	val = readl(priv->base + LDDCKR);
	if (priv->dot_clk && (((val >> 16) & 0x7) == icksel))
		return 0;
	if (priv->dot_clk)
		pm_runtime_disable(&pdev->dev);

	priv->dot_clk_source = icksel;

	if (clkstr) {
		priv->dot_clk = clk_get(NULL, clkstr);
		if (IS_ERR(priv->dot_clk)) {
			dev_err(&pdev->dev, "cannot get dot clock %s\n",
				clkstr);
			return PTR_ERR(priv->dot_clk);
		}
	}
	atomic_set(&priv->hw_usecnt, -1);

	/* Runtime PM support involves two step for this driver:
	 * 1) Enable Runtime PM
	 * 2) Force Runtime PM Resume since hardware is accessed from probe()
	 */
	pm_runtime_enable(&pdev->dev);
	pm_runtime_resume(&pdev->dev);
	return 0;
}

static int sh_mobile_lcdc_set_mode(struct fb_info *info)
{
	struct sh_mobile_lcdc_priv *priv = info->par;
	struct fb_var_screeninfo *var = &info->var;
	struct sh_videomode_ext *ext = sh_find_videomode_ext(info);
	unsigned long ext_flags = 0;
	u32 ldmt1r, ldddsr, lddfr;

	trace();

	/* signal configration */
	if (ext)
		ext_flags = ext->flags;
	ldmt1r = 0;
	ldmt1r |= (var->sync & FB_SYNC_VERT_HIGH_ACT) ? 0 : 1 << 28;
	ldmt1r |= (var->sync & FB_SYNC_HOR_HIGH_ACT) ? 0 : 1 << 27;
	ldmt1r |= (ext_flags & LCDC_FLAGS_DWPOL) ? 1 << 26 : 0;
	ldmt1r |= (ext_flags & LCDC_FLAGS_DIPOL) ? 1 << 25 : 0;
	ldmt1r |= (ext_flags & LCDC_FLAGS_DAPOL) ? 1 << 24 : 0;
	ldmt1r |= (ext_flags & LCDC_FLAGS_HSCNT) ? 1 << 17 : 0;
	ldmt1r |= (ext_flags & LCDC_FLAGS_DWCNT) ? 1 << 16 : 0;

	switch (var->nonstd) {
	case SH_FIXUP_YUV420:
		ldmt1r |= 0xb;
		ldddsr = 0x00000007;
		lddfr = 0x00030000;
		var->red.offset = 0;
		var->red.length = 0;
		var->green.offset = 0;
		var->green.length = 0;
		var->blue.offset = 0;
		var->blue.length = 0;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	case SH_FIXUP_RGB888:
		ldmt1r |= 0xb;
		ldddsr = 0x00000007;
		lddfr = 0x0000000b;
		var->red.offset = 16;
		var->red.length = 8;
		var->green.offset = 8;
		var->green.length = 8;
		var->blue.offset = 0;
		var->blue.length = 8;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	case SH_FIXUP_XRGB8888:
	default:
		ldmt1r |= 0xb;
		ldddsr = 0x00000004;
		lddfr = 0x00000000;
		var->red.offset = 16;
		var->red.length = 8;
		var->green.offset = 8;
		var->green.length = 8;
		var->blue.offset = 0;
		var->blue.length = 8;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	}
	var->red.msb_right = 0;
	var->green.msb_right = 0;
	var->blue.msb_right = 0;
	var->transp.msb_right = 0;

	s_writel_both(ldmt1r, priv->base + LDMT1R);
	s_writel_both(0x00000000, priv->base + LDMT2R);
	s_writel_both(0x00000000, priv->base + LDMT3R);
	s_writel(ldddsr, priv->base + LDDDSR);
	s_writel_both(lddfr, priv->base + LDDFR);

	return 0;
}

static int sh_mobile_lcdc_set_timing(struct fb_info *info,
				     struct fb_var_screeninfo *var)
{
	struct sh_mobile_lcdc_priv *priv = info->par;
	struct sh_mobile_lcdc_info *pdata = priv_to_pdata(priv);
	void *base = priv->base;
	int clock_source;
	u32 divm, divn;
	u32 val;
	trace();

	d_dbg("      pixclock = %d\n", var->pixclock);
	d_dbg("          xres = %d\n", var->xres);
	d_dbg("          yres = %d\n", var->yres);
	d_dbg("   left_margin = %d\n", var->left_margin);
	d_dbg("  right_margin = %d\n", var->right_margin);
	d_dbg("     hsync_len = %d\n", var->hsync_len);
	d_dbg("  upper_margin = %d\n", var->upper_margin);
	d_dbg("  lower_margin = %d\n", var->lower_margin);
	d_dbg("     vsync_len = %d\n", var->vsync_len);

	/* horizontal configuration */
	val = (var->xres + var->hsync_len +
	       var->left_margin + var->right_margin) / 8;
	val |= (var->xres / 8) << 16;
	s_writel_both(val, base + LDHCNR);

	val = (var->xres + var->right_margin) / 8;
	val |= (var->hsync_len / 8) << 16;
	s_writel_both(val, base + LDHSYNR);

	/* vertical configuration */
	val = (var->yres + var->vsync_len +
	       var->upper_margin + var->lower_margin);
	val |= var->yres << 16;
	s_writel_both(val, base + LDVLNR);

	val = var->yres + var->lower_margin;
	val |= var->vsync_len << 16;
	s_writel_both(val, base + LDVSYNR);

	/* configure clocks */
	if (var->xres == 720 && var->yres == 480)
		clock_source = LCDC_CLK_DV;
	else
		clock_source = pdata->clock_source;

	sh_mobile_lcdc_start_stop(priv, 0);
	sh_mobile_lcdc_clk_off(priv);
	sh_mobile_lcdc_setup_clocks(priv, clock_source);
	sh_mobile_lcdc_clk_on(priv);

	divm = (clk_get_rate(priv->dot_clk) / 1000) /
		(1000000000 / var->pixclock);
	if (divm == 1) {
		divm = 1 << 6;
		divn = 1;
	} else {
		divn = (1 << (divm / 2)) - 1;
	}

	writel(0x00000000, base + LDDCKPAT1R);
	writel(divn, base + LDDCKPAT2R);

	val = divm;
	val |= (priv->dot_clk_source << 16);
	writel(val, base + LDDCKR);

	sh_mobile_lcdc_start_stop(priv, 1);
	return 0;
}

static int sh_mobile_lcdc_map_smem(struct fb_info *info,
				   struct fb_fix_screeninfo *fix)
{
	struct sh_mobile_lcdc_priv *priv = info->par;
	struct fb_var_screeninfo *var = &info->var;
	u32 line_length;

	trace();

	info->fix.smem_len = fix->smem_len;
	info->fix.line_length = fix->line_length;
	info->var.yres_virtual = info->var.yres * 2;
	info->var.xres_virtual = info->var.xres;

	info->screen_size = info->fix.smem_len;
	info->screen_base = dma_alloc_coherent(&priv->pdev->dev,
					       info->fix.smem_len,
					       (dma_addr_t *)&info->fix.smem_start,
					       GFP_KERNEL);
	if (!info->screen_base) {
		dev_err(&priv->pdev->dev, "unable to allocate buffer\n");
		return -ENOMEM;
	}

	if (is_yuv_mode(info->var.nonstd)) {
		memset(info->screen_base, 0x80, info->fix.smem_len);
		memset(info->screen_base, 0x00, info->var.xres*info->var.yres);
		memset(info->screen_base + (var->yoffset*info->fix.line_length),
		       0x00, info->var.xres*info->var.yres);
	} else {
		memset(info->screen_base, 0x00, info->fix.smem_len);
	}

	line_length = info->fix.line_length;

	if (is_yuv_mode(info->var.nonstd))
		line_length /= 2;

	if (var->vmode & FB_VMODE_INTERLACED)
		line_length *= 2;

	s_writel_both(line_length, priv->base + LDMLSR);
	s_writel_seta(info->fix.smem_start, priv->base + LDSA1R);
	if (is_yuv_mode(info->var.nonstd))
		s_writel_seta(info->fix.smem_start +
			      info->var.xres * info->var.yres,
			      priv->base + LDSA2R);
	if (var->vmode & FB_VMODE_INTERLACED) {
		s_writel_setb(info->fix.smem_start + info->fix.line_length,
			      priv->base + LDSA1R);
		if (is_yuv_mode(info->var.nonstd))
			s_writel_setb(info->fix.smem_start +
				      (info->var.xres + 1) * info->var.yres,
				      priv->base + LDSA2R);
		s_writel(0x00000001, priv->base + LDRCNTR);
	} else {
		s_writel(0x00000000, priv->base + LDRCNTR);
	}

	return 0;
}

static int sh_mobile_lcdc_unmap_smem(struct fb_info *info)
{
	struct sh_mobile_lcdc_priv *priv = info->par;

	trace();

	dma_free_coherent(&priv->pdev->dev, info->fix.smem_len,
			  info->screen_base, (dma_addr_t)info->fix.smem_start);
	info->fix.smem_len = 0;

	return 0;
}

static int sh_mobile_fb_pan_display(struct fb_var_screeninfo *var,
				     struct fb_info *info)
{
	struct sh_mobile_lcdc_priv *priv = info->par;
	unsigned long cur_pan_offset;
	unsigned long new_pan_offset;

	trace();

	cur_pan_offset = readl(priv->base + LDSA1R) - info->fix.smem_start;
	new_pan_offset = (var->yoffset * info->fix.line_length) +
		(var->xoffset * (info->var.bits_per_pixel / 8));

	if (new_pan_offset == cur_pan_offset)
		return 0;	/* No change, do nothing */

	s_writel_seta(info->fix.smem_start + new_pan_offset,
		      priv->base + LDSA1R);
	s_writel_seta(info->fix.smem_start + new_pan_offset +
		      info->var.xres * info->var.yres,
		      priv->base + LDSA2R);
	if (info->var.vmode & FB_VMODE_INTERLACED) {
		s_writel_setb(info->fix.smem_start + info->fix.line_length +
			      new_pan_offset, priv->base + LDSA1R);
		s_writel_setb(info->fix.smem_start + new_pan_offset +
			      info->var.xres * (info->var.yres + 1),
			      priv->base + LDSA2R);
	}

	return 0;
}

static int sh_mobile_lcdc_setcolreg(u_int regno,
				    u_int red, u_int green, u_int blue,
				    u_int transp, struct fb_info *info)
{
	u32 *palette = info->pseudo_palette;

	trace();

	if (regno >= PALETTE_NR)
		return -EINVAL;

	/* only FB_VISUAL_TRUECOLOR supported */
	red >>= 16 - info->var.red.length;
	green >>= 16 - info->var.green.length;
	blue >>= 16 - info->var.blue.length;
	transp >>= 16 - info->var.transp.length;

	palette[regno] = ((red << info->var.red.offset) |
			  (green << info->var.green.offset) |
			  (blue << info->var.blue.offset) |
			  (transp << info->var.transp.offset));

	return 0;
}

static int sh_mobile_lcdc_check_var(struct fb_var_screeninfo *var,
				    struct fb_info *info)
{
	trace();

	/* for Android */
	return 0;
}

static int sh_mobile_lcdc_set_par(struct fb_info *info)
{
	struct fb_fix_screeninfo fix;
	struct sh_videomode_ext *ext = sh_find_videomode_ext(info);
	int ret;

	trace();

	ret = sh_mobile_lcdc_set_mode(info);
	if (ret)
		return ret;

	ret = sh_mobile_lcdc_set_timing(info, &info->var);
	if (ret)
		return ret;

	/* update video memory */
	fix.line_length = info->var.xres * (info->var.bits_per_pixel / 8);
	fix.smem_len = roundup(fix.line_length * info->var.yres,
			       lcm(fix.line_length, 0x1000));
	fix.smem_len *= 2;

	if (info->fix.smem_len < fix.smem_len) {
		if (info->fix.smem_start)
			sh_mobile_lcdc_unmap_smem(info);

		sh_mobile_lcdc_map_smem(info, &fix);
	}

	/* for Android */
	info->fix.ypanstep = 1;

	info->fix.type = FB_TYPE_PACKED_PIXELS;
	info->fix.visual = FB_VISUAL_TRUECOLOR;
	info->fix.accel = FB_ACCEL_NONE;

	if (ext && ext->fixup)
		ret = ext->fixup(info, info->var.nonstd);

	return ret;
}

static int  sh_mobile_lcdc_blank(int blank, struct fb_info *info)
{
	struct sh_mobile_lcdc_priv *priv = info->par;
	u32 ldcnt2r = readl(priv->base + LDCNT2R);

	switch (blank) {
	case FB_BLANK_POWERDOWN:
	case FB_BLANK_VSYNC_SUSPEND:
	case FB_BLANK_HSYNC_SUSPEND:
	case FB_BLANK_NORMAL:
		ldcnt2r &= ~0x01;
		break;
	case FB_BLANK_UNBLANK:
		ldcnt2r |= 0x01;
		break;
	}
	writel(ldcnt2r, priv->base + LDCNT2R);

	return 0;
}

static struct fb_ops sh_mobile_lcdc_ops = {
	.owner          = THIS_MODULE,
	.fb_setcolreg	= sh_mobile_lcdc_setcolreg,
	.fb_pan_display = sh_mobile_fb_pan_display,
	.fb_check_var	= sh_mobile_lcdc_check_var,
	.fb_set_par	= sh_mobile_lcdc_set_par,
	.fb_blank	= sh_mobile_lcdc_blank,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
};

static void sh_mobile_lcdc_reset(struct fb_info *info)
{
	struct sh_mobile_lcdc_priv *priv = info->par;
	u32 val;

	val = readl(priv->base + LDCNT2R);
	val |= LCDC_RESET;
	writel(val, priv->base + LDCNT2R);
	while (readl(priv->base + LDCNT2R) & LCDC_RESET)
		cpu_relax();

	writel(0x2, priv->base + LDCNT2R);
	writel(0x1, priv->base + LDCNT1R);

	sh_mobile_lcdc_start_stop(priv, 0);
}

static int __devinit sh_mobile_lcdc_probe(struct platform_device *pdev)
{
	struct fb_info *info;
	struct sh_mobile_lcdc_priv *priv;
	struct sh_mobile_lcdc_info *pdata;
	struct resource *res;
	int device_id = pdev->id;
	char *args;
	char *setup_videomode;
	char *default_videomode = NULL;
	char *label = "rmobile_lcdcfb.0";
	u32 bits_per_pixel;
	int irq;
	int ret;

	trace();

	if (!pdev->dev.platform_data) {
		dev_err(&pdev->dev, "no platform data defined\n");
		return  -EINVAL;
	}

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	irq = platform_get_irq(pdev, 0);
	if (!res || irq < 0) {
		dev_err(&pdev->dev, "cannot get platform resources\n");
		return -ENOENT;
	}

	info = framebuffer_alloc(sizeof(struct sh_mobile_lcdc_priv),
				 &pdev->dev);
	if (!info) {
		dev_err(&pdev->dev, "unable to allocate fb_info\n");
		return -ENOMEM;
	}

	priv = info->par;
	priv->info = info;
	priv->pdev = pdev;
	priv->base = ioremap_nocache(res->start, (res->end - res->start) + 1);
	info->fbops = &sh_mobile_lcdc_ops;
	platform_set_drvdata(pdev, priv);
	pdata = priv_to_pdata(priv);

	ret = sh_mobile_lcdc_setup_clocks(priv,  pdata->clock_source);
	if (ret) {
		dev_err(&pdev->dev, "unable to setup clocks\n");
		goto unmap_resource;
	}
	sh_mobile_lcdc_clk_on(priv);

	sprintf(label, "rmobile_lcdcfb.%d", device_id);
	setup_videomode = (char *)pdata->ch[0].default_mode;
	fb_get_options(label, &args);
	if (args)
		setup_videomode = args;
	if (videomode)
		setup_videomode = videomode;
	bits_per_pixel = pdata->ch[0].bpp;
	{
		char *opt;
		while ((opt = strsep(&setup_videomode, ",")) != NULL) {
			if (!*opt)
				continue;
			if (!strncasecmp(opt, "rgb888", 6)) {
				info->var.nonstd = SH_FIXUP_RGB888;
				bits_per_pixel = 24;
			} else if (!strncasecmp(opt, "xrgb8888", 8)) {
				info->var.nonstd = SH_FIXUP_XRGB8888;
				bits_per_pixel = 32;
			} else if (!strncasecmp(opt, "yuv420", 6)) {
				info->var.nonstd = SH_FIXUP_YUV420;
				bits_per_pixel = 16;
			} else
				default_videomode = opt;
		}
	}

	fb_videomode_to_modelist(pdata->ch[0].modedb, pdata->ch[0].modedb_size,
				 &info->modelist);
	ret = fb_find_mode(&info->var, info,
			   default_videomode,
			   pdata->ch[0].modedb, pdata->ch[0].modedb_size,
			   NULL, bits_per_pixel);
	if (ret == 0) {
		ret = -EINVAL;
		goto unmap_resource;
	} else if (ret == 1)
		dev_info(&pdev->dev, "using \"%s\" videomode.\n",
			 default_videomode);
	else if (ret == 3)
		dev_info(&pdev->dev, "using default videomode.\n");

	sh_mobile_lcdc_reset(info);
	sh_mobile_lcdc_set_par(info);

	info->pseudo_palette = &priv->pseudo_palette;
	info->flags = FBINFO_FLAG_DEFAULT;
	info->var.activate = FB_ACTIVATE_NOW;
	ret = fb_alloc_cmap(&info->cmap, 16, 0);
	if (ret < 0)
		goto unmap_resource;

	ret = register_framebuffer(info);
	if (ret < 0)
		goto dealloc_cmap;

	dev_info(&pdev->dev, "registered %s as %dx%d %dbpp.\n", pdev->name,
		 info->var.xres, info->var.yres, info->var.bits_per_pixel);

	return 0;

dealloc_cmap:
	fb_dealloc_cmap(&info->cmap);
unmap_resource:
	iounmap(priv->base);
	framebuffer_release(info);
	return ret;
}

static int sh_mobile_lcdc_remove(struct platform_device *pdev)
{
	struct sh_mobile_lcdc_priv *priv = platform_get_drvdata(pdev);
	struct fb_info *info = priv->info;

	unregister_framebuffer(info);
	sh_mobile_lcdc_start_stop(priv, 0);
	sh_mobile_lcdc_clk_off(priv);
	sh_mobile_lcdc_unmap_smem(info);

	fb_dealloc_cmap(&info->cmap);
	iounmap(priv->base);
	framebuffer_release(info);

	pm_runtime_disable(&pdev->dev);

	return 0;
}

static int sh_mobile_lcdc_suspend(struct device *dev)
{
	trace();

	/* todo */

	return 0;
}

static int sh_mobile_lcdc_resume(struct device *dev)
{
	trace();

	/* todo */

	return 0;
}

static int sh_mobile_lcdc_runtime_suspend(struct device *dev)
{
	trace();

	/* todo */

	return 0;
}

static int sh_mobile_lcdc_runtime_resume(struct device *dev)
{
	trace();

	/* todo */

	return 0;
}

static const struct dev_pm_ops sh_mobile_lcdc_dev_pm_ops = {
	.suspend = sh_mobile_lcdc_suspend,
	.resume = sh_mobile_lcdc_resume,
	.runtime_suspend = sh_mobile_lcdc_runtime_suspend,
	.runtime_resume = sh_mobile_lcdc_runtime_resume,
};

static struct platform_driver sh_mobile_lcdc_driver = {
	.driver		= {
		.name		= "sh_mobile_lcdc_fb",
		.owner		= THIS_MODULE,
		.pm		= &sh_mobile_lcdc_dev_pm_ops,
	},
	.probe		= sh_mobile_lcdc_probe,
	.remove		= sh_mobile_lcdc_remove,
};

static int __init sh_mobile_lcdc_init(void)
{
	return platform_driver_register(&sh_mobile_lcdc_driver);
}

static void __exit sh_mobile_lcdc_exit(void)
{
	platform_driver_unregister(&sh_mobile_lcdc_driver);
}

module_param(videomode, charp, 0);
module_init(sh_mobile_lcdc_init);
module_exit(sh_mobile_lcdc_exit);

MODULE_DESCRIPTION("R-Mobile A1 LCDC Framebuffer driver");
MODULE_AUTHOR("Atmark Techno, Inc.");
MODULE_LICENSE("GPL v2");
