/*
 * Watchdog driver for Renesas SH-Mobile processors.
 *
 * Copyright (C) 2011 Renesas Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/err.h>
#include <linux/platform_device.h>
#include <linux/moduleparam.h>
#include <linux/clk.h>
#include <linux/bitops.h>
#include <linux/io.h>
#include <linux/uaccess.h>
#include <linux/slab.h>

#define RWDTCNT		0x00
#define RWDTCSRA	0x04
#define RWDTCSRB	0x08

#define WATCHDOG_TIMEOUT 30		/* 30 sec default timeout */
static int timeout = WATCHDOG_TIMEOUT;
module_param(timeout, int, 0);
MODULE_PARM_DESC(timeout,
	"Watchdog timeout in seconds. (1 <= timeout <= 3600, default="
				__MODULE_STRING(WATCHDOG_TIMEOUT) ")");

static int nowayout = WATCHDOG_NOWAYOUT;
module_param(nowayout, int, 0);
MODULE_PARM_DESC(nowayout,
	"Watchdog cannot be stopped once started (default="
				__MODULE_STRING(WATCHDOG_NOWAYOUT) ")");

static DEFINE_SPINLOCK(wdt_lock);

static void __iomem	*wdt_base;
static struct clk	*wdt_clk;
static struct resource	*wdt_irq;
static int		wdt_enable;
static unsigned long	wdt_users;

static u8 wdt_read(int offset)
{
	return __raw_readb(wdt_base + offset);
}

static void wdt_write(int offset, u8 val)
{
	u16 set_val = val;

	switch (offset) {
	case RWDTCNT:
		__raw_writew(set_val | 0x5a00, wdt_base + offset);
		break;
	case RWDTCSRA:
	case RWDTCSRB:
		__raw_writew(set_val | 0xa500, wdt_base + offset);
		break;
	}
}

static irqreturn_t sh_mobile_wdt_irq(int irq, void *arg)
{
	printk(KERN_CRIT "Watchdog Timer expired. Reboot ignored.\n");
	return IRQ_HANDLED;
}

static void sh_mobile_wdt_ping(void)
{
	spin_lock(&wdt_lock);
	if (!(wdt_read(RWDTCSRA) & 0x20))
		wdt_write(RWDTCNT, 0);
	spin_unlock(&wdt_lock);
}

static void sh_mobile_wdt_enable(void)
{
	u8 cks30 = 0x06;
	u8 cks54 = 0x03;
	int tmp;

	spin_lock(&wdt_lock);
	if (!wdt_enable) {
		clk_enable(wdt_clk);
		wdt_enable = 1;
	}

	/* Setting timeout */
	wdt_write(RWDTCSRA, 0x0f);
	if (timeout < 10) {
		cks54 = 0;
		cks30 = timeout ? timeout : 1;
	} else if (timeout < 60) {
		tmp = timeout / 5;
		cks54 = 1;
		cks30 = (timeout % 5) ? tmp + 1 : tmp;
		timeout = cks30 * 5;
	} else if (timeout < 60*10) {
		tmp = timeout / 60;
		cks54 = 2;
		cks30 = (timeout % 60) ? tmp + 1 : tmp;
		timeout = cks30 * 60;
	} else if (timeout < 60*30) {
		tmp = timeout / (60*5);
		cks54 = 2;
		cks30 = (timeout % (60*5)) ? tmp + 1 : tmp;
		timeout = cks30 * 60 * 5;
	} else
		timeout = cks30 * 60 * 5;
	wdt_write(RWDTCSRB, (cks54 << 4) | cks30);

	/* Enable timer */
	wdt_write(RWDTCSRA, 0x8f);
	spin_unlock(&wdt_lock);
}

static void sh_mobile_wdt_disable(void)
{
	spin_lock(&wdt_lock);
	if (!wdt_enable) {
		clk_enable(wdt_clk);
		wdt_enable = 1;
	}

	wdt_write(RWDTCSRA, 0);
	wdt_write(RWDTCNT, 0);

	if (!wdt_enable) {
		clk_disable(wdt_clk);
		wdt_enable = 0;
	}
	spin_unlock(&wdt_lock);
}

static int sh_mobile_wdt_open(struct inode *inode, struct file *file)
{
	if (test_and_set_bit(1, &wdt_users))
		return -EBUSY;

	sh_mobile_wdt_enable();

	return nonseekable_open(inode, file);
}

static int sh_mobile_wdt_release(struct inode *inode, struct file *file)
{
#ifndef CONFIG_WATCHDOG_NOWAYOUT
	sh_mobile_wdt_disable();
#endif
	wdt_users = 0;
	return 0;
}

static ssize_t sh_mobile_wdt_write(struct file *file, const char __user *data,
		size_t len, loff_t *ppos)
{
	if (len)
		sh_mobile_wdt_ping();
	return len;
}

static const struct watchdog_info ident = {
	.options	= WDIOF_SETTIMEOUT | WDIOF_KEEPALIVEPING,
	.identity	= "SH-Mobile Watchdog",
};

static long sh_mobile_wdt_ioctl(struct file *file, unsigned int cmd,
				unsigned long arg)
{
	switch (cmd) {
	case WDIOC_GETSUPPORT:
		return copy_to_user((struct watchdog_info __user *)arg, &ident,
				sizeof(ident));
	case WDIOC_GETSTATUS:
	case WDIOC_GETBOOTSTATUS:
		return put_user(0, (int *)arg);
	case WDIOC_KEEPALIVE:
		sh_mobile_wdt_ping();
		return 0;
	case WDIOC_SETTIMEOUT:
		spin_lock(&wdt_lock);
		if (get_user(timeout, (int __user *)arg)) {
			spin_unlock(&wdt_lock);
			return -EFAULT;
		}
		spin_unlock(&wdt_lock);
		sh_mobile_wdt_disable();
		sh_mobile_wdt_enable();
		/* Fall through */
	case WDIOC_GETTIMEOUT:
		return put_user(timeout, (int __user *)arg);
	default:
		return -ENOTTY;
	}
}

static const struct file_operations sh_mobile_wdt_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.write		= sh_mobile_wdt_write,
	.unlocked_ioctl	= sh_mobile_wdt_ioctl,
	.open		= sh_mobile_wdt_open,
	.release	= sh_mobile_wdt_release,
};

static struct miscdevice sh_mobile_wdt_miscdev = {
	.minor = WATCHDOG_MINOR,
	.name = "watchdog",
	.fops = &sh_mobile_wdt_fops,
};

static int __devinit sh_mobile_wdt_probe(struct platform_device *pdev)
{
	struct resource *res, *mem;
	int ret;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		dev_err(&pdev->dev, "failed to get memory region resource\n");
		return -ENOENT;
	}

	mem = request_mem_region(res->start, resource_size(res), pdev->name);
	if (!mem) {
		dev_err(&pdev->dev, "failed to get memory region\n");
		return -ENOENT;
	}

	wdt_clk = clk_get(&pdev->dev, "rwdt0");
	if (IS_ERR(wdt_clk)) {
		ret = PTR_ERR(wdt_clk);
		goto err1;
	}

	wdt_base = ioremap(res->start, resource_size(res));
	if (!wdt_base) {
		ret = -ENOMEM;
		goto err2;
	}

	wdt_users = 0;
	wdt_enable = 0;

	sh_mobile_wdt_disable();

	wdt_irq = platform_get_resource(pdev, IORESOURCE_IRQ, 0);
	if (wdt_irq) {
		ret = request_irq(wdt_irq->start, sh_mobile_wdt_irq, 0,
				dev_name(&pdev->dev), NULL);
		if (ret)
			goto err3;
	}

	ret = misc_register(&sh_mobile_wdt_miscdev);
	if (ret)
		goto err4;

	return 0;

err4:
	if (wdt_irq)
		free_irq(wdt_irq->start, NULL);
err3:
	iounmap(wdt_base);
err2:
	clk_put(wdt_clk);
err1:
	release_mem_region(res->start, resource_size(res));
	return ret;
}

static int __devexit sh_mobile_wdt_remove(struct platform_device *pdev)
{
	struct resource *res = platform_get_resource(pdev, IORESOURCE_MEM, 0);

	if (!res)
		return -ENOENT;

	if (wdt_irq)
		free_irq(wdt_irq->start, NULL);

	sh_mobile_wdt_disable();

	misc_deregister(&sh_mobile_wdt_miscdev);
	release_mem_region(res->start, resource_size(res));

	clk_put(wdt_clk);
	iounmap(wdt_base);

	return 0;
}

static void sh_mobile_wdt_shutdown(struct platform_device *pdev)
{
	if (wdt_users)
		sh_mobile_wdt_disable();
}

static struct platform_driver sh_mobile_wdt_driver = {
	.probe		= sh_mobile_wdt_probe,
	.remove		= __devexit_p(sh_mobile_wdt_remove),
	.shutdown	= sh_mobile_wdt_shutdown,
	.driver = {
		.name = "sh_mobile_wdt",
		.owner	= THIS_MODULE,
	},
};

static int __init sh_mobile_wdt_init(void)
{
	return platform_driver_register(&sh_mobile_wdt_driver);
}

static void __exit sh_mobile_wdt_exit(void)
{
	platform_driver_unregister(&sh_mobile_wdt_driver);
}

module_init(sh_mobile_wdt_init);
module_exit(sh_mobile_wdt_exit);

MODULE_AUTHOR("Renesas");
MODULE_LICENSE("GPL");
MODULE_ALIAS_MISCDEV(WATCHDOG_MINOR);
MODULE_ALIAS("platform:watchdog");
