/*
 * FIFO-incorporating Melody and Serial Interface (FMSI) support
 *
 * Copyright (C) 2011 Renesas Electronics Corporation.
 *
 * Copyright (C) 2009 Renesas Solutions Corp.
 * Kuninori Morimoto <morimoto.kuninori@renesas.com>
 *
 * Based on ssi.c
 * Copyright (c) 2007 Manuel Lauss <mano@roarinelk.homelinux.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/list.h>
#include <linux/pm_runtime.h>
#include <linux/io.h>
#include <linux/slab.h>
#include <linux/clk.h>
#include <linux/string.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/initval.h>
#include <sound/soc.h>
#include <sound/pcm_params.h>


#include "sh_fmsi.h"
#include "./sosi/include/RSPF0_API_S0.h"
#include "./sosi/include/RSPF0_FMSIDRV_API.h"

#ifdef CONFIG_SND_SOC_SH4_FMSI_SRC
#include "fmsi_src.h"
#endif

/* #define FMSI_DEBUG_0 */
/* #define FMSI_DEBUG_1 */

#ifdef FMSI_DEBUG_0
	#define dbg_msg_0(FMT, ARGS...) \
		printk(KERN_INFO "[Info]%s : "FMT, __func__, ##ARGS)
	#define dbg_msg_1(FMT, ARGS...) \
		printk(KERN_INFO "[Info]%s : "FMT, __func__, ##ARGS)
#else
	#ifdef FMSI_DEBUG_1
		#define dbg_msg_0(FMT, ARGS...) {}
		#define dbg_msg_1(FMT, ARGS...) \
			printk(KERN_INFO "[Info]%s : "FMT, __func__, ##ARGS)
	#else
		#define dbg_msg_0(FMT, ARGS...) {}
		#define dbg_msg_1(FMT, ARGS...) {}
	#endif
#endif

#define S_DO_FMT	0x0000
#define S_DOFF_CTL	0x0004
#define S_DOFF_ST	0x0008
#define S_DI_FMT	0x000C
#define S_DIFF_CTL	0x0010
#define S_DIFF_ST	0x0014
#define S_DIDT		0x0020
#define S_DODT		0x0024
#define S_MUTE_ST	0x0028

#define SOMI_CKG_RW_MD	0x01C0
#define SOMI_CKG_RW_RV	0x01C4
#define SIMI_CKG_RW_MD	0x01C8
#define SIMI_CKG_RW_RV	0x01CC
#define CKG_SMD		0x01D0
#define TMR_CTL		0x01D8
#define TMR_CLR		0x01DC

#define INT_ST		0x0200
#define IEMSK		0x0204
#define IMSK		0x0208
#define MUTE		0x020C
#define CLK_RST		0x0210
#define SOFT_RST	0x0214
#define FIFO_SZ		0x0218
#define REG_END		FIFO_SZ


/* Bit Info. */
/* S_DO_FMT */
/* S_DI_FMT */
#define CR_FMT(param) ((param) << 4)
# define CR_MONO	0x0
# define CR_MONO_D	0x1
# define CR_PCM		0x2
# define CR_I2S		0x3
# define CR_TDM		0x4
# define CR_TDM_D	0x5


/* S_DOFF_CTL */
/* S_DIFF_CTL */


/* S_DOFF_ST */
/* S_DIFF_ST */


/* S_DIDT */
/* S_DODT */

/* S_MUTE_ST */
/* SOMI_CKG_RW_MD */
/* SOMI_CKG_RW_RV */
/* SIMI_CKG_RW_MD */
/* SIMI_CKG_RW_RV */

/* CKG_SMD */

/* TMR_CTL */
/* TMR_CLR */

/* INT_ST */
#define INT_IN	(1 << 4)
#define INT_OUT	(1 << 0)

/* IEMSK */

/* IMSK */

/* MUTE */

/* CLK_RST */

/* SOFT_RS */

/* FIFO_SZ */



/* DOFF_CTL */
/* DIFF_CTL */
#define IRQ_HALF	0x00100000
#define FIFO_CLR	0x00000001

/* DOFF_ST */
#define ERR_OVER	0x00000010
#define ERR_UNDER	0x00000001
#define ST_ERR		(ERR_OVER | ERR_UNDER)

/* CLK_RST */
#define B_CLK		0x00000010
#define A_CLK		0x00000001


/* SOFT_RST */
#define PBSR		(1 << 12) /* Port B Software Reset */
#define PASR		(1 <<  8) /* Port A Software Reset */
#define IR		(1 <<  4) /* Interrupt Reset */
#define FSISR		(1 <<  0) /* Software Reset */

/* FIFO_SZ */
#define OUT_SZ_MASK	0x7
#define BO_SZ_SHIFT	8
#define AO_SZ_SHIFT	0


#define FMSI_RATES SNDRV_PCM_RATE_8000_96000

#define FMSI_FMTS (SNDRV_PCM_FMTBIT_S24_LE | SNDRV_PCM_FMTBIT_S16_LE)

#define SPU2_DSP0		0
#define SPU2_DSP1		1
#define SPU2_USE_DSP_NUM	SPU2_DSP1
#define SPU2_ASYNC		0
#define SPU2_SYNC		1

void __iomem *spu2_base;

static DECLARE_MUTEX(sem);
static DECLARE_WAIT_QUEUE_HEAD(wait);
static long rspf0_event;

#include "fmsi_spu_init.c"


/************************************************************************


		struct


************************************************************************/
struct fmsi_priv {
	void __iomem *base;
	struct fmsi_master *master;

	struct snd_pcm_substream	*src_substream;
	int				src_period_len;
	int				src_periods;
	int				src_over_period;
	int				src_data_status;
	unsigned long			src_index;
	unsigned long			src_total_size;
	RSPF0_Path			src_g_path;

	struct snd_pcm_substream	*dst_substream;
	int				dst_period_len;
	int				dst_periods;
	int				dst_over_period;
	int				dst_data_status;
	unsigned long			dst_index;
	unsigned long			dst_total_size;
	RSPF0_Path			dst_g_path;

	unsigned long src_cb_size;
	unsigned long dst_cb_size;
};

struct fmsi_regs {
	u32 int_st;
	u32 iemsk;
	u32 imsk;
};

struct fmsi_master {
	void __iomem *base;
	int irq[2];
	struct fmsi_priv fmsi;
	struct fmsi_regs *regs;
	struct sh_fmsi_platform_info *info;

	RSPF0_FMSIDRV_Port *port_src;
	RSPF0_FMSIDRV_Port *port_dst;

	spinlock_t lock;
};

/* User-created function */
static void fmsi_rspf0_cb_spu_read
	(unsigned long *addr, unsigned long *data, unsigned long size);
static void fmsi_rspf0_cb_spu_write
	(unsigned long *addr, unsigned long *data, unsigned long size);
static long fmsi_rspf0_cb_wait_event(unsigned long timeout);
static long fmsi_rspf0_cb_set_event(void);
static long fmsi_rspf0_cb_leave_critical_section(void);
static long fmsi_rspf0_cb_enter_critical_section(void);

static void fmsi_read
	(unsigned long *addr, unsigned long *data, unsigned long size);
static void fmsi_write
	(unsigned long *addr, unsigned long *data, unsigned long size);
static void fmsi_wait(unsigned long usec);

static void fmsi_rspf0_cb_source_end_callback
		(RSPF0_Path *path, unsigned long size);
static void fmsi_rspf0_cb_dest_end_callback
		(RSPF0_Path *path, unsigned long size, unsigned long flush);
static void fmsi_rspf0_src_cb_error_callback
		(RSPF0_Path *path, unsigned long info);
static void fmsi_rspf0_dst_cb_error_callback
		(RSPF0_Path *path, unsigned long info);

static void source_transfer_start(struct snd_pcm_substream *substream);
static void dest_transfer_start(struct snd_pcm_substream *substream);


/************************************************************************


		SPU2 Init Quit function


************************************************************************/
static int fmsi_spu_sound_mw_init(struct fmsi_master *master)
{
	RSPF0_Func func;

	RSPF0_FMSIDRV_InitParam param;

	RSPF0_FMSIDRV_Clock clock_system;
	unsigned long direction;

	RSPF0_FMSIDRV_PortParam port_param;

	RSPF0_FMSIDRV_IoParam io_param;

	int ret;

	dbg_msg_0("fmsi_spu_sound_mw_init\n");

	/* SPU_Init() */
	ret = SPU_Init(master);
	if (ret) {
		printk(KERN_ERR "SPU_Init err\n");
		return -1;
	}

	/* RSPF0_FMSIDRV_Init() */
	param.address_offset = 0x001c0400;
	param.read =	fmsi_read;
	param.write =	fmsi_write;
	param.wait =	fmsi_wait;
	ret = RSPF0_FMSIDRV_Init(&param);
	if (ret != RSPF0_FMSIDRV_RTN_SUCCESS) {
		printk(KERN_ERR "SRSPF0_FMSIDRV_Init err\n");
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_Init ok....\n");



	/* RSPF0_FMSIDRV_OpenPort for src */
	clock_system.master_mode =	RSPF0_FMSIDRV_MASTERCLOCK_ONE;
	clock_system.slave_mode =	RSPF0_FMSIDRV_SLAVECLOCK_ONE;
	direction =			RSPF0_FMSIDRV_DIRECTION_SOUTPUT;
	ret = RSPF0_FMSIDRV_OpenPort(&clock_system, direction, &master->port_src);
	if (ret != RSPF0_FMSIDRV_RTN_SUCCESS) {
		printk(KERN_ERR "RSPF0_FMSIDRV_OpenPort for src err\n");
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_OpenPort for src ok....\n");

	/* RSPF0_FMSIDRV_OpenPort for dst */
	clock_system.master_mode =	RSPF0_FMSIDRV_MASTERCLOCK_ONE;
	clock_system.slave_mode =	RSPF0_FMSIDRV_SLAVECLOCK_ONE;
	direction =			RSPF0_FMSIDRV_DIRECTION_SINPUT;
	ret = RSPF0_FMSIDRV_OpenPort(&clock_system, direction, &master->port_dst);
	if (ret != RSPF0_FMSIDRV_RTN_SUCCESS) {
		printk(KERN_ERR "RSPF0_FMSIDRV_OpenPort for dst err\n");
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_OpenPort for dst ok....\n");



	/* RSPF0_FMSIDRV_SetPortParam for src */
	port_param.audio_clock_mode =		FMSI_AUDIOCLK;
	port_param.bit_per_fs_mode =		FMSI_BITPERFS;
	port_param.master.lrclk_inverse =	RSPF0_FMSIDRV_POLARITY_NORMAL;
	port_param.master.bclk_inverse =	RSPF0_FMSIDRV_POLARITY_REVERSE;
	port_param.slave.lrclk_inverse =	RSPF0_FMSIDRV_POLARITY_NORMAL;
	port_param.slave.bclk_inverse =		RSPF0_FMSIDRV_POLARITY_REVERSE;
	ret = RSPF0_FMSIDRV_SetPortParam(master->port_src, &port_param);
	if (ret != RSPF0_FMSIDRV_RTN_SUCCESS) {
		printk(KERN_ERR "RSPF0_FMSIDRV_SetPortParam for src err\n");
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_SetPortParam for src ok....\n");

	/* RSPF0_FMSIDRV_SetPortParam for dst */
	port_param.audio_clock_mode =		FMSI_AUDIOCLK;
	port_param.bit_per_fs_mode =		FMSI_BITPERFS;
	port_param.master.lrclk_inverse =	RSPF0_FMSIDRV_POLARITY_NORMAL;
	port_param.master.bclk_inverse =	RSPF0_FMSIDRV_POLARITY_REVERSE;
	port_param.slave.lrclk_inverse =	RSPF0_FMSIDRV_POLARITY_NORMAL;
	port_param.slave.bclk_inverse =		RSPF0_FMSIDRV_POLARITY_REVERSE;
	ret = RSPF0_FMSIDRV_SetPortParam(master->port_dst, &port_param);
	if (ret != RSPF0_FMSIDRV_RTN_SUCCESS) {
		printk(KERN_ERR "RSPF0_FMSIDRV_SetPortParam for dst err\n");
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_SetPortParam for dst ok....\n");



	/* RSPF0_FMSIDRV_SetIoParam for src */
	io_param.bws =			RSPF0_FMSIDRV_BWS_16BIT;
	io_param.format =		RSPF0_FMSIDRV_FORMAT_I2S;
	io_param.tdm_channel =		RSPF0_FMSIDRV_TDM_CHANNEL_1;
#ifdef CONFIG_SND_SOC_SH4_FMSI_MASTER
	io_param.msmode =		RSPF0_FMSIDRV_MSMODE_MASTER;
#else
	io_param.msmode =		RSPF0_FMSIDRV_MSMODE_SLAVE;
#endif
	ret = RSPF0_FMSIDRV_SetIoParam(master->port_src, &io_param);
	if (ret != RSPF0_FMSIDRV_RTN_SUCCESS) {
		printk(KERN_ERR "RSPF0_FMSIDRV_SetIoParam for src err\n");
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_SetIoParam for src ok....\n");

	/* RSPF0_FMSIDRV_SetIoParam for dst */
	io_param.bws =			RSPF0_FMSIDRV_BWS_16BIT;
	io_param.format =		RSPF0_FMSIDRV_FORMAT_I2S;
	io_param.tdm_channel =		RSPF0_FMSIDRV_TDM_CHANNEL_1;
#ifdef CONFIG_SND_SOC_SH4_FMSI_MASTER
	io_param.msmode =		RSPF0_FMSIDRV_MSMODE_MASTER;
#else
	io_param.msmode =		RSPF0_FMSIDRV_MSMODE_SLAVE;
#endif
	ret = RSPF0_FMSIDRV_SetIoParam(master->port_dst, &io_param);
	if (ret != RSPF0_FMSIDRV_RTN_SUCCESS) {
		printk(KERN_ERR "RSPF0_FMSIDRV_SetIoParam for dst err\n");
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_SetIoParam for dst ok....\n");



	/* SPU_InitDSP() */
	ret = SPU_InitDSP(spu2_base,
		SPU2_USE_DSP_NUM, RSPF0_PBANK_NUM, RSPF0_XBANK_NUM);

	if (ret) {
		printk(KERN_ERR "SPU_InitDSP err\n");
		return -1;
	}



	enable_irq(master->irq[SPU2_USE_DSP_NUM]);
	dbg_msg_0("SPU_InitDSP ok....\n");



	/* RSPF0_Init() */
	func.reg_read               = fmsi_rspf0_cb_spu_read;
	func.reg_write              = fmsi_rspf0_cb_spu_write;
	func.wait_for_event         = fmsi_rspf0_cb_wait_event;
	func.set_event              = fmsi_rspf0_cb_set_event;
	func.enter_critical_section = fmsi_rspf0_cb_enter_critical_section;
	func.leave_critical_section = fmsi_rspf0_cb_leave_critical_section;

	ret = RSPF0_Init(SPU2_USE_DSP_NUM, &func, RSPF0_ADDRESS_MODE_32BIT);
	if (ret != RSPF0_RTN_SUCCESS) {
		printk(KERN_ERR "RSPF0_Init err\n");
		return -1;
	}
	dbg_msg_0("RSPF0_Init ok....\n");



	/* RSPF0_FMSIDRV_SoftwareReset() */
	ret = RSPF0_FMSIDRV_SoftwareReset();
	if (ret != RSPF0_FMSIDRV_RTN_SUCCESS) {
		printk(KERN_ERR "RSPF0_FMSIDRV_SoftwareReset err\n");
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_SoftwareReset ok....\n");



	/* RSPF0_FMSIDRV_ResetPort() for src */
	ret = RSPF0_FMSIDRV_ResetPort(master->port_src);
	if (ret != RSPF0_FMSIDRV_RTN_SUCCESS) {
		printk(KERN_ERR "RSPF0_FMSIDRV_ResetPort for src err\n");
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_ResetPort for src ok....\n");

	/* RSPF0_FMSIDRV_ResetPort() for dst */
	ret = RSPF0_FMSIDRV_ResetPort(master->port_dst);
	if (ret != RSPF0_FMSIDRV_RTN_SUCCESS) {
		printk(KERN_ERR "RSPF0_FMSIDRV_ResetPort for dst err\n");
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_ResetPort for dst ok....\n");



	/* RSPF0_FMSIDRV_StartPort() for src */
	ret = RSPF0_FMSIDRV_StartPort(master->port_src);
	if (ret != RSPF0_FMSIDRV_RTN_SUCCESS) {
		printk(KERN_ERR "RSPF0_FMSIDRV_StartPort for src err\n");
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_StartPort for src ok....\n");

	/* RSPF0_FMSIDRV_StartPort() for dst */
	ret = RSPF0_FMSIDRV_StartPort(master->port_dst);
	if (ret != RSPF0_FMSIDRV_RTN_SUCCESS) {
		printk(KERN_ERR "RSPF0_FMSIDRV_StartPort for dsterr\n");
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_StartPort for dst ok....\n");



	return 0;
}

static int fmsi_spu_sound_mw_quit(struct fmsi_master *master)
{
	int ret;

	/* SPU_QuitDSP() */
	ret = SPU_QuitDSP(spu2_base, SPU2_USE_DSP_NUM);
	if (ret != 0) {
		printk(KERN_ERR "SPU_QuitDSP err\n");
		return -1;
	}
	dbg_msg_0("SPU_QuitDSP ok....\n");

	disable_irq(master->irq[SPU2_USE_DSP_NUM]);

	/* RSPF0_FMSIDRV_ResetPort() for src */
	ret = RSPF0_FMSIDRV_ResetPort(master->port_src);
	if (ret != 0) {
		printk(KERN_ERR "RSPF0_FMSIDRV_ResetPort for src err(%d)\n", ret);
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_ResetPort for src ok....\n");

	/* RSPF0_FMSIDRV_ResetPort() for dst */
	ret = RSPF0_FMSIDRV_ResetPort(master->port_dst);
	if (ret != 0) {
		printk(KERN_ERR "RSPF0_FMSIDRV_ResetPort for dst err(%d)\n", ret);
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_ResetPort for dst ok....\n");


	/* RSPF0_FMSIDRV_ClosePort() for src */
	ret = RSPF0_FMSIDRV_ClosePort(master->port_src);
	if (ret != 0) {
		printk(KERN_ERR "RSPF0_FMSIDRV_ClosePort for src err(%d)\n", ret);
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_ClosePort for src ok....\n");

	/* RSPF0_FMSIDRV_ClosePort() for dst */
	ret = RSPF0_FMSIDRV_ClosePort(master->port_dst);
	if (ret != 0) {
		printk(KERN_ERR "RSPF0_FMSIDRV_ClosePort for dst err(%d)\n", ret);
		return -1;
	}
	dbg_msg_0("RSPF0_FMSIDRV_ClosePort for dst ok....\n");



	/* SPU_Quit() */
	ret = SPU_Quit(spu2_base);
	if (ret != 0) {
		printk(KERN_ERR "SPU_Quit err\n");
		return -1;
	}
	dbg_msg_0("SPU_Quit ok....\n");

	return 0;
}

/************************************************************************


		basic read write function


************************************************************************/
static void __fmsi_reg_write(u32 reg, u32 data)
{
	/* valid data area is 24bit */
	data &= 0x00ffffff;

	dbg_msg_0("__fmsi_reg_write [%x]<<%x\n", reg, data);
	RSPF0_IoWrite(SPU2_USE_DSP_NUM, reg, data, SPU2_ASYNC);

	return;
}

static u32 __fmsi_reg_read(u32 reg)
{
	unsigned long val;
	u32 ret;

	ret = RSPF0_IoRead(SPU2_USE_DSP_NUM, reg, &val);
	dbg_msg_0("__fmsi_reg_read [%x]>>%x\n", reg, val);

	return val;
}

/************************************************************************


		basic function


************************************************************************/
static struct fmsi_master *fmsi_get_master(struct fmsi_priv *fmsi)
{
	return fmsi->master;
}

static struct snd_soc_dai *fmsi_get_dai(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;

	return  machine->cpu_dai;
}

static struct fmsi_priv *fmsi_get_priv(struct snd_pcm_substream *substream)
{
	struct snd_soc_dai *dai = fmsi_get_dai(substream);

	return dai->private_data;
}

static void fmsi_stream_push(struct fmsi_priv *fmsi,
			    struct snd_pcm_substream *substream,
			    u32 buffer_len,
			    u32 period_len)
{
	int is_play = substream->stream == SNDRV_PCM_STREAM_PLAYBACK;

	if (is_play) {	/* Source */
		fmsi->src_substream	= substream;
		fmsi->src_period_len	= period_len;
		fmsi->src_periods	= 0;
		fmsi->src_total_size	= buffer_len;
		fmsi->src_index		= 0;
	} else {	/* Dest */
		fmsi->dst_substream	= substream;
		fmsi->dst_period_len	= period_len;
		fmsi->dst_periods	= 0;
		fmsi->dst_total_size	= buffer_len;
		fmsi->dst_index		= 0;
	}

	return;
}

static void fmsi_stream_pop(struct fmsi_priv *fmsi,
			    struct snd_pcm_substream *substream)
{
	int is_play = substream->stream == SNDRV_PCM_STREAM_PLAYBACK;

	if (is_play) {	/* Source */
		fmsi->src_substream	= NULL;
		fmsi->src_period_len	= 0;
		fmsi->src_periods	= 0;
		fmsi->src_total_size	= 0;
		fmsi->src_index		= 0;
	} else {	/* Dest */
		fmsi->dst_substream	= NULL;
		fmsi->dst_period_len	= 0;
		fmsi->dst_periods	= 0;
		fmsi->dst_total_size	= 0;
		fmsi->dst_index		= 0;
	}

	return;
}

/************************************************************************


		irq function


************************************************************************/

/************************************************************************


		ctrl function


************************************************************************/
/* playback interrupt */
static int fmsi_data_push(struct snd_pcm_substream *substream, int startup)
{
	struct fmsi_master *master;
	master = fmsi_get_master(fmsi_soc_dai.private_data);

	dbg_msg_0("fmsi_data_push\n");
	source_transfer_start(substream);
	dbg_msg_0("fmsi_data_push end\n");

	return 0;
}

static int fmsi_data_pop(struct snd_pcm_substream *substream, int startup)
{
	struct fmsi_master *master;
	master = fmsi_get_master(fmsi_soc_dai.private_data);

	dbg_msg_0("fmsi_data_pop\n");

	dest_transfer_start(substream);

	dbg_msg_0("fmsi_data_pop end\n");

	return 0;
}

/*---------------------------*/
/*   User define functions   */
/*---------------------------*/


static long fmsi_rspf0_cb_enter_critical_section(void)
{
	dbg_msg_0("fmsi_rspf0_cb_enter_critical_section\n");

	if (down_interruptible(&sem)) {
		dbg_msg_0(KERN_ERR "down error\n");
		return -1;
	}

	return 0;
}

static long fmsi_rspf0_cb_leave_critical_section(void)
{
	dbg_msg_0("fmsi_rspf0_cb_leave_critical_section\n");
	up(&sem);

	return 0;
}

static long fmsi_rspf0_cb_set_event(void)
{
	/* refer to the SPU2 middleware porting guide */

	dbg_msg_0("fmsi_rspf0_cb_set_event\n");

	rspf0_event = 1;
	wake_up_interruptible(&wait);

	return 0;
}

static long fmsi_rspf0_cb_wait_event(unsigned long timeout)
{
	long val;

	dbg_msg_0("fmsi_rspf0_cb_wait_event\n");

	val = wait_event_interruptible_timeout(wait, rspf0_event != 0,
							timeout * HZ / 1000);
	rspf0_event = 0;

	if (val == 0) {
		dbg_msg_1("fmsi_rspf0_cb_wait_event TO\n");
		return -1;
	}

	dbg_msg_0("fmsi_rspf0_cb_wait_event end\n");

	return 0;
}

static void fmsi_rspf0_cb_spu_read(unsigned long *addr,
				unsigned long *data, unsigned long size)
{
	/* refer to the SPU2 middleware porting guide */
	unsigned long i;

	dbg_msg_0("fmsi_rspf0_cb_spu_read 0x%x+0x%x - 0x%x+0x%x ",
			spu2_base, (unsigned int)addr,
			spu2_base, (unsigned int)(addr + size));

	for (i = 0; i < size; i++)
		data[i] = readl(spu2_base + (unsigned long)(addr + i));

	return;
}

static void fmsi_rspf0_cb_spu_write(unsigned long *addr,
				unsigned long *data, unsigned long size)
{
	/* refer to the SPU2 middleware porting guide */
	unsigned long i;

	dbg_msg_0("fmsi_rspf0_cb_spu_write 0x%x+0x%x - 0x%x+0x%x ",
			spu2_base, (unsigned int)addr,
			spu2_base, (unsigned int)(addr + size));

	for (i = 0; i < size; i++)
		writel(data[i], spu2_base +  (unsigned long)(addr + i));

	return;
}


static void sio_pcm_kick_src(unsigned long data)
{
	struct snd_pcm_runtime *runtime;
	struct snd_pcm_substream *substream = NULL;
	struct fmsi_priv *fmsi =
		container_of((RSPF0_Path *)data, struct fmsi_priv, src_g_path);

	unsigned char *start;
	unsigned long tr_size;
	unsigned long tr_f_size;
	int ret;
	unsigned long size = fmsi->src_cb_size;

	struct fmsi_master *master;
	master = fmsi_get_master(fmsi_soc_dai.private_data);

	if (fmsi->src_substream == NULL) {
		dbg_msg_1("substream was cleared\n");
		return;
	}

	substream	= fmsi->src_substream;
	runtime		= substream->runtime;
	start		= (unsigned char *)runtime->dma_addr;

	dbg_msg_0("fmsi_rspf0_cb_source_end_callback\n");

	dbg_msg_1("sent src_index:%lx\n", fmsi->src_index);
	dbg_msg_1("     size     :%lx Sample\n", size);
	fmsi->src_index += (size * 2);

	if (fmsi->src_index >=
	    fmsi->src_period_len * (fmsi->src_periods + 1)) {
		dbg_msg_1("<elapsed period>\n");
		fmsi->src_over_period = 1;
		fmsi->src_periods = (fmsi->src_periods + 1) % runtime->periods;

		if (fmsi->src_periods == 0) {
			dbg_msg_1("<buffer end>\n");
			fmsi->src_index = 0;
		}
	}

	tr_size = fmsi->src_period_len - (fmsi->src_period_len % 32);
	if (tr_size + fmsi->src_index > fmsi->src_total_size)
		tr_size = fmsi->src_total_size - fmsi->src_index;

	tr_f_size = tr_size / 2;

	if (fmsi->src_data_status == 0) {
		fmsi->src_over_period = 0;
		ret  = RSPF0_FmsiStart(&(fmsi->src_g_path),
						RSPF0_SYNC_NO_RESPONSE);
		dbg_msg_0("RSPF0_FmsiStart ret(%d)\n", ret);
		fmsi->src_data_status = 1;
	}

	dbg_msg_1("send src_index:%lx\n", fmsi->src_index);
	dbg_msg_1("     tr_size  :%lx Byte\n", tr_size);
	dbg_msg_1("     tr_f_size:%lx Sample\n", tr_f_size);
	ret = RSPF0_SourceTransferStart(&(fmsi->src_g_path),
				(short *)&(start[fmsi->src_index]), tr_f_size);

	if (ret)
		printk(KERN_ERR "RSPF0_SourceTransferStart err (%d)\n", ret);

	if (fmsi->src_over_period) {
		fmsi->src_over_period = 0;
		dbg_msg_1("<snd_pcm_period_elapsed>\n");
		snd_pcm_period_elapsed(substream);
	}

	dbg_msg_0("transfer data end (%d)\n", ret);

	return;
}

static void sio_pcm_kick_dst(unsigned long data)
{
	struct snd_pcm_substream *substream;
	struct snd_pcm_runtime *runtime;
	struct fmsi_priv *fmsi =
		container_of((RSPF0_Path *)data, struct fmsi_priv, dst_g_path);

	unsigned long tr_size;
	unsigned long tr_f_size;
	unsigned char *start;
	int ret;
	unsigned long size = fmsi->dst_cb_size;

	struct fmsi_master *master;
	master = fmsi_get_master(fmsi_soc_dai.private_data);

	if (fmsi->dst_substream == NULL) {
		dbg_msg_1("substream was cleared\n");
		return;
	}

	substream	= fmsi->dst_substream;
	runtime		= substream->runtime;
	start		= (unsigned char *)runtime->dma_addr;

	dbg_msg_1("got dst_index:%lx\n", fmsi->dst_index);
	dbg_msg_1("    size     :%lx Sample\n", size);
	fmsi->dst_index += (size * 2);

	if (fmsi->dst_index >=
	    fmsi->dst_period_len * (fmsi->dst_periods + 1)) {
		dbg_msg_1("<elapsed period>\n");
		fmsi->dst_over_period = 1;
		fmsi->dst_periods = (fmsi->dst_periods + 1) % runtime->periods;

		if (fmsi->dst_periods == 0) {
			dbg_msg_1("<buffer end>\n");
			fmsi->dst_index = 0;
		}
	}

	tr_size = fmsi->dst_period_len - (fmsi->dst_period_len % 64);
	if (tr_size + fmsi->dst_index > fmsi->dst_total_size)
		tr_size = fmsi->dst_total_size - fmsi->dst_index;

	tr_f_size = tr_size / 2;

	if (fmsi->dst_data_status == 0) {
		fmsi->dst_over_period = 0;
		ret  = RSPF0_FmsiStart(&(fmsi->dst_g_path),
						RSPF0_SYNC_NO_RESPONSE);
		dbg_msg_0("RSPF0_FmsiStart ret(%d)\n", ret);
		fmsi->dst_data_status = 1;
	}

	dbg_msg_1("get dst_index:%lx\n", fmsi->dst_index);
	dbg_msg_1("    tr_size   :%lx Byte\n", tr_size);
	dbg_msg_1("    tr_f_size :%lx Sample\n", tr_f_size);

	ret = RSPF0_DestTransferStart(&(fmsi->dst_g_path),
				(short *)&(start[fmsi->dst_index]), tr_f_size);

	if (ret)
		printk(KERN_ERR "RSPF0_DestTransferStart err (%d)\n", ret);

	if (fmsi->dst_over_period) {
		fmsi->dst_over_period = 0;
		dbg_msg_1("<snd_pcm_period_elapsed>\n");
		snd_pcm_period_elapsed(substream);
	}

	return;
}


DECLARE_TASKLET(fmsi_source_end_callback_tasklet, sio_pcm_kick_src, 0);
DECLARE_TASKLET(fmsi_dest_end_callback_tasklet, sio_pcm_kick_dst, 0);

/*-------------------------------------------*/
/*   User-created function & test function   */
/*-------------------------------------------*/
/* User-created function for sound */
/* Transfer end callback(Source) */
static void fmsi_rspf0_cb_source_end_callback(RSPF0_Path *path,
						unsigned long size)
{
	struct fmsi_priv *fmsi =
		container_of(path, struct fmsi_priv, src_g_path);

	struct fmsi_master *master;
	master = fmsi_get_master(fmsi_soc_dai.private_data);

	if (fmsi->src_substream == NULL) {
		dbg_msg_1("substream was cleared\n");
		return;
	}

	fmsi->src_cb_size = size;

	fmsi_source_end_callback_tasklet.data = (unsigned long)path;
	tasklet_schedule(&fmsi_source_end_callback_tasklet);

	return;
}

/* Transfer end callback(Dest) */
static void fmsi_rspf0_cb_dest_end_callback(RSPF0_Path *path,
				unsigned long size, unsigned long flush)
{
	struct snd_pcm_substream *substream;
	struct snd_pcm_runtime *runtime;
	struct fmsi_priv *fmsi =
		container_of(path, struct fmsi_priv, dst_g_path);

	unsigned char *start;

	struct fmsi_master *master;
	master = fmsi_get_master(fmsi_soc_dai.private_data);

	if (fmsi->dst_substream == NULL) {
		dbg_msg_1("substream was cleared\n");
		return;
	}

	substream	= fmsi->dst_substream;
	runtime		= substream->runtime;
	start		= (unsigned char *)runtime->dma_addr;

	if (flush) {
		dbg_msg_1("fmsi_rspf0_cb_dest_end_callback [flush]\n");
		return;
	}

	fmsi->dst_cb_size = size;

	fmsi_dest_end_callback_tasklet.data = (unsigned long)path;
	tasklet_schedule(&fmsi_dest_end_callback_tasklet);

	return;
}


/* error */
static void fmsi_rspf0_src_cb_error_callback(RSPF0_Path *path,
							unsigned long info)
{
	int ret;

	printk(KERN_ERR "fmsi_rspf0_src_cb_error_callback err (%ld)\n", info);

	ret = RSPF0_FmsiStop(path, 0, 0);
	if (ret != RSPF0_RTN_SUCCESS)
		printk(KERN_ERR "RSPF0_FmsiStop for src err (%d)\n", ret);

	ret = RSPF0_FmsiStart(path, 0);

	if (ret != RSPF0_RTN_SUCCESS)
		printk(KERN_ERR "RSPF0_FmsiStart for src err (%d)\n", ret);

	return;
}

static void fmsi_rspf0_dst_cb_error_callback(RSPF0_Path *path,
							unsigned long info)
{
	int ret;

	printk(KERN_ERR "fmsi_rspf0_dst_cb_error_callback err (%ld)\n", info);

	ret = RSPF0_FmsiStop(path, 0, 0);
	if (ret != RSPF0_RTN_SUCCESS)
		printk(KERN_ERR "RSPF0_FmsiStop for src err (%d)\n", ret);

	ret = RSPF0_FmsiStart(path, 0);

	if (ret != RSPF0_RTN_SUCCESS)
		printk(KERN_ERR "RSPF0_FmsiStart for src err (%d)\n", ret);

	return;
}

static void fmsi_read(unsigned long *addr, unsigned long *data,
							unsigned long size)
{
	int i;
	dbg_msg_0("fmsi_read  addr= 0x%x\n", (unsigned int)addr);

	for (i = 0 ; i < size ; i++)
		data[i] = __fmsi_reg_read(0xFE200000 +
						(unsigned long)(addr + i));

	return;
}

static void fmsi_write(unsigned long *addr, unsigned long *data,
							unsigned long size)
{
	int i;
	dbg_msg_0("fmsi_write addr = 0x%x\n", (unsigned int)addr);

	for (i = 0; i < size; i++)
		__fmsi_reg_write(0xFE200000 + (u32)(addr + i), data[i]);

	return;
}

static void fmsi_wait(unsigned long usec)
{

	dbg_msg_0("fmsi_wait\n");
	udelay(usec);

	return;
}

/* Start Source data */
static void source_transfer_start(struct snd_pcm_substream *substream)
{
	struct fmsi_priv *fmsi = fmsi_get_priv(substream);
	struct snd_pcm_runtime *runtime;

	unsigned long tr_size;
	unsigned long tr_f_size;
	unsigned char *start;
	int ret;

	struct fmsi_master *master;
	master = fmsi_get_master(fmsi_soc_dai.private_data);

	dbg_msg_0("source_transfer_start\n");

	runtime		= substream->runtime;
	start		= (unsigned char *)runtime->dma_addr;

	dbg_msg_1("start         :%hx\n", start);
	dbg_msg_1("src_period_len    :%lx Byte\n", fmsi->src_period_len);
	dbg_msg_1("src_periods       :%x\n", runtime->periods);
	dbg_msg_1("src_total_size:%lx Byte\n", fmsi->src_total_size);

	tr_size = fmsi->src_period_len - (fmsi->src_period_len % 32);
	if (tr_size + fmsi->src_index > fmsi->src_total_size)
		tr_size = fmsi->src_total_size - fmsi->src_index;

	tr_f_size = tr_size / 2;

	dbg_msg_0("transfer data %x+%x -> %x+%x\n",
		(unsigned int)start,
		(unsigned int)fmsi->src_index,
		(unsigned int)start,
		(unsigned int)(fmsi->src_index + tr_size));

	if (fmsi->src_data_status == 0) {
		fmsi->src_over_period = 0;
		ret  = RSPF0_FmsiStart(&(fmsi->src_g_path),
					RSPF0_SYNC_NO_RESPONSE);
		dbg_msg_0("RSPF0_FmsiStart ret(%d)\n", ret);
		fmsi->src_data_status = 1;
	}

	dbg_msg_1("send src_index:%x\n", fmsi->src_index);
	dbg_msg_1("     tr_size  :%x Byte\n", tr_size);
	dbg_msg_1("     tr_f_size:%x Sample\n", tr_f_size);
	ret = RSPF0_SourceTransferStart(&(fmsi->src_g_path),
			(short *)&(start[fmsi->src_index]), tr_f_size);

	if (ret)
		printk(KERN_ERR "RSPF0_SourceTransferStart err (%d)\n", ret);

	dbg_msg_0("transfer data end (%d)\n", ret);

	return;
}

/* Start Dest data */
static void dest_transfer_start(struct snd_pcm_substream *substream)
{
	struct fmsi_priv *fmsi = fmsi_get_priv(substream);
	struct snd_pcm_runtime *runtime;

	u8 *start;
	unsigned long tr_size;
	unsigned long tr_f_size;
	int ret;

	struct fmsi_master *master;
	master = fmsi_get_master(fmsi_soc_dai.private_data);

	dbg_msg_1("dest_transfer_start\n");

	if (!fmsi		||
	    !fmsi->dst_substream	||
	    !fmsi->dst_substream->runtime) {
		printk(KERN_ERR "dest_transfer_start err (%d)\n", -EINVAL);
		return ;
	}

	runtime		= substream->runtime;
	start		= (unsigned char *)runtime->dma_addr;

	dbg_msg_1("start         :%hx\n", start);
	dbg_msg_1("dst_period_len    :%lx Byte\n", fmsi->dst_period_len);
	dbg_msg_1("dst_periods       :%x\n", runtime->periods);
	dbg_msg_1("dst_total_size:%lx Byte\n", fmsi->dst_total_size);

	tr_size = fmsi->dst_period_len - (fmsi->dst_period_len % 64);
	if (tr_size + fmsi->dst_index > fmsi->dst_total_size)
		tr_size = fmsi->dst_total_size - fmsi->dst_index;

	tr_f_size = tr_size / 2;

	if (fmsi->dst_data_status == 0) {
		fmsi->dst_over_period = 0;
		ret  = RSPF0_FmsiStart(&(fmsi->dst_g_path),
					RSPF0_SYNC_NO_RESPONSE);
		dbg_msg_1("RSPF0_FmsiStart ret(%d)\n", ret);
		fmsi->dst_data_status = 1;
	}

	dbg_msg_1("get dst_index:%x\n", fmsi->dst_index);
	dbg_msg_1("    tr_size  :%x Byte\n", tr_size);
	dbg_msg_1("    tr_f_size:%x Sample\n", tr_f_size);
	ret = RSPF0_DestTransferStart(&(fmsi->dst_g_path),
			(short *)&(start[fmsi->dst_index]), tr_f_size);
	if (ret)
		printk(KERN_ERR "RSPF0_DestTransferStart err (%d)\n", ret);

	dbg_msg_0("dest data end (%d)\n", ret);

	return;
}

/************************************************************************


		dai ops


************************************************************************/
static struct snd_soc_dai_ops fmsi_dai_ops = {
};

/************************************************************************


		pcm ops


************************************************************************/
static struct snd_pcm_hardware fmsi_pcm_hardware = {
	.info =		SNDRV_PCM_INFO_INTERLEAVED	|
			SNDRV_PCM_INFO_MMAP		|
			SNDRV_PCM_INFO_MMAP_VALID	|
			SNDRV_PCM_INFO_PAUSE,
	.formats		= FMSI_FMTS,
	.rates			= FMSI_RATES,
	.rate_min		= 8000,
	.rate_max		= 192000,
	.channels_min		= 1,
	.channels_max		= 2,
	.buffer_bytes_max	= 64 * 1024,
	.period_bytes_min	= 32,
	.period_bytes_max	= 8192,
	.periods_min		= 1,
	.periods_max		= 32,
	.fifo_size		= 256,
};

static int fmsi_pcm_open(struct snd_pcm_substream *substream)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	int ret = 0;

	RSPF0_OpenParam open_param;
	struct fmsi_priv *fmsi = fmsi_get_priv(substream);
	int is_play = substream->stream == SNDRV_PCM_STREAM_PLAYBACK;

	dbg_msg_1("fmsi_pcm_open\n");

	rspf0_event		= 0;

	/* RSPF0_Open() */
	open_param.swap_mode           = RSPF0_SWAP_MODE_WORD;

	open_param.channel             = RSPF0_CHANNEL_STEREO;
	open_param.sound.fir_mode      = RSPF0_FIR_OFF;

	if (is_play) {
		fmsi->src_data_status	= 0;
		dbg_msg_1("fmsi_pcm_open (source)\n");
		open_param.fmsi_error_callback  =
					fmsi_rspf0_src_cb_error_callback;
		open_param.source_port         = RSPF0_PORT_CPU;
		open_param.dest_port           = RSPF0_PORT_FMSI_A;
		open_param.source_end_callback =
					fmsi_rspf0_cb_source_end_callback;
		open_param.dest_end_callback   = NULL;
#ifdef CONFIG_SND_SOC_SH4_FMSI_SRC
		open_param.sound.fir_mode      = RSPF0_FIR_SRC;
#endif
		ret = RSPF0_Open(SPU2_USE_DSP_NUM,
					&(fmsi->src_g_path), &open_param);
	} else {
		fmsi->dst_data_status	= 0;
		dbg_msg_1("fmsi_pcm_open (dest)\n");
		open_param.fmsi_error_callback  =
					fmsi_rspf0_dst_cb_error_callback;
		open_param.source_port         = RSPF0_PORT_FMSI_A;
		open_param.dest_port           = RSPF0_PORT_CPU;
		open_param.source_end_callback = NULL;
		open_param.dest_end_callback   =
					fmsi_rspf0_cb_dest_end_callback;
		ret = RSPF0_Open(SPU2_USE_DSP_NUM,
					&(fmsi->dst_g_path), &open_param);
	}

	if (ret != RSPF0_RTN_SUCCESS) {
		printk(KERN_ERR "RSPF0_Open err (%d)\n", ret);
		return -1;
	}

	dbg_msg_0("RSPF0_Open ok....\n");

	snd_soc_set_runtime_hwparams(substream, &fmsi_pcm_hardware);

	ret = snd_pcm_hw_constraint_integer(runtime,
					    SNDRV_PCM_HW_PARAM_PERIODS);

	return ret;
}

static int fmsi_pcm_close(struct snd_pcm_substream *substream)
{
	struct fmsi_priv *fmsi = fmsi_get_priv(substream);
	int is_play = substream->stream == SNDRV_PCM_STREAM_PLAYBACK;
	int ret;

	dbg_msg_1("fmsi_pcm_close\n");

	if (is_play) {	/* Source */
		ret = RSPF0_FmsiStop(&(fmsi->src_g_path), 0,
						RSPF0_SYNC_NO_RESPONSE);
		dbg_msg_0("RSPF0_FmsiStop for src\n");
		if (ret != RSPF0_RTN_SUCCESS)
			printk(KERN_ERR "RSPF0_FmsiStop for src err (%d)\n",
									ret);

		ret = RSPF0_Clear(&(fmsi->src_g_path));
		dbg_msg_0("RSPF0_Clear for src\n");
		if (ret != RSPF0_RTN_SUCCESS)
			printk(KERN_ERR "RSPF0_Clear for src err (%d)\n",
									ret);

		ret = RSPF0_Close(&(fmsi->src_g_path));
		dbg_msg_0("RSPF0_Close for src\n");
		if (ret != RSPF0_RTN_SUCCESS)
			printk(KERN_ERR "RSPF0_Close for src err (%d)\n",
									ret);

	} else {	/* Dest */
		ret = RSPF0_FmsiStop(&(fmsi->dst_g_path), 0,
						RSPF0_SYNC_NO_RESPONSE);
		dbg_msg_0("RSPF0_FmsiStop for dst\n");
		if (ret != RSPF0_RTN_SUCCESS)
			printk(KERN_ERR "RSPF0_FmsiStop for dst err (%d)\n",
									ret);

		ret = RSPF0_Clear(&(fmsi->dst_g_path));
		dbg_msg_0("RSPF0_Clear for dst\n");
		if (ret != RSPF0_RTN_SUCCESS)
			printk(KERN_ERR "RSPF0_Clear for dst err (%d)\n",
									ret);

		ret = RSPF0_Close(&(fmsi->dst_g_path));
		dbg_msg_0("RSPF0_Close for dst\n");
		if (ret != RSPF0_RTN_SUCCESS)
			printk(KERN_ERR "RSPF0_Close for dst err (%d)\n",
									ret);
	}

	if (ret != RSPF0_RTN_SUCCESS)
		return -1;

	return 0;
}

static int fmsi_hw_params(struct snd_pcm_substream *substream,
			 struct snd_pcm_hw_params *hw_params)
{
#ifdef CONFIG_SND_SOC_SH4_FMSI_SRC
	struct fmsi_priv *fmsi = fmsi_get_priv(substream);
	int is_play = substream->stream == SNDRV_PCM_STREAM_PLAYBACK;
	unsigned long ratio;
	int ret;
#endif
	dbg_msg_1("fmsi_hw_params\n");

#ifdef CONFIG_SND_SOC_SH4_FMSI_SRC
	if (is_play) {
		switch (params_rate(hw_params)) {
		case 8000:
			ratio = SRC_RATIO_INPUT_8000;
			break;
		case 11025:
			ratio = SRC_RATIO_INPUT_11025;
			break;
		case 12000:
			ratio = SRC_RATIO_INPUT_12000;
			break;
		case 14700:
			ratio = SRC_RATIO_INPUT_14700;
			break;
		case 16000:
			ratio = SRC_RATIO_INPUT_16000;
			break;
		case 22050:
			ratio = SRC_RATIO_INPUT_22050;
			break;
		case 24000:
			ratio = SRC_RATIO_INPUT_24000;
			break;
		case 29400:
			ratio = SRC_RATIO_INPUT_29400;
			break;
		case 32000:
			ratio = SRC_RATIO_INPUT_32000;
			break;
		case 44100:
			ratio = SRC_RATIO_INPUT_44100;
			break;
		case 48000:
			ratio = SRC_RATIO_INPUT_48000;
			break;
		default:
			printk(KERN_ERR "input rate err\n");
			return -EINVAL;
			break;
		}

		ret = RSPF0_SetSrc(&(fmsi->src_g_path), ratio,
			RSPF0_SYNC_WAIT_RESPONSE);

		if (ret != RSPF0_RTN_SUCCESS) {
			if ( ret == RSPF0_RTN_ERROR_SPU_API)
				printk(KERN_WARNING
					"RSPF0_SetSrc warning (%d)\n", ret);
			
			else {
				printk(KERN_ERR
					"RSPF0_SetSrc err (%d)\n", ret);
				return -1;
			}
		}
	}
#endif
	return snd_pcm_lib_malloc_pages(substream,
					params_buffer_bytes(hw_params));
}

static int fmsi_hw_free(struct snd_pcm_substream *substream)
{

	dbg_msg_1("fmsi_hw_free\n");

	return snd_pcm_lib_free_pages(substream);
}

static int fmsi_pcm_trigger(struct snd_pcm_substream *substream, int cmd)
{
	struct fmsi_priv *fmsi = fmsi_get_priv(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	int is_play = substream->stream == SNDRV_PCM_STREAM_PLAYBACK;
	int ret = 0;

	if (is_play) {	/* Source */
		fmsi->src_substream = substream;
	} else {	/* Dest */
		fmsi->dst_substream = substream;
	}

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		dbg_msg_1("fmsi_pcm_trigger (START)\n");
		if (is_play) {	/* Source */
			fmsi_stream_push(fmsi, substream,
				frames_to_bytes(runtime, runtime->buffer_size),
				frames_to_bytes(runtime, runtime->period_size));
			fmsi_data_push(substream, 1);
		} else {	/* Dest */
			fmsi_stream_push(fmsi, substream,
				frames_to_bytes(runtime, runtime->buffer_size),
				frames_to_bytes(runtime, runtime->period_size));
			fmsi_data_pop(substream, 1);
		}
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		dbg_msg_1("fmsi_pcm_trigger (STOP)\n");
		fmsi_stream_pop(fmsi, substream);
		break;
	}

	return ret;
}

static snd_pcm_uframes_t fmsi_pointer(struct snd_pcm_substream *substream)
{
	struct fmsi_priv *fmsi = fmsi_get_priv(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	int is_play = substream->stream == SNDRV_PCM_STREAM_PLAYBACK;
	long location;
	unsigned long tr_f_size;

	if (is_play) {	/* Source */
		dbg_msg_1("fmsi_pointer:%lx tr_f_size:%lx Sample\n",
						fmsi->src_index, tr_f_size);
		location = fmsi->src_index;
	} else {	/* Dest */
		dbg_msg_1("fmsi_pointer:%lx tr_f_size:%lx Sample\n",
						fmsi->dst_index, tr_f_size);
		location = fmsi->dst_index;
	}
	if (location < 0)
		location = 0;

	tr_f_size = bytes_to_frames(runtime, location);

	return tr_f_size;
}

static struct snd_pcm_ops fmsi_pcm_ops = {
	.open		= fmsi_pcm_open,
	.close          = fmsi_pcm_close,
	.ioctl		= snd_pcm_lib_ioctl,
	.hw_params	= fmsi_hw_params,
	.hw_free	= fmsi_hw_free,
	.trigger	= fmsi_pcm_trigger,
	.pointer	= fmsi_pointer,
};

/************************************************************************


		snd_soc_platform


************************************************************************/
#define PREALLOC_BUFFER		(32 * 1024)
#define PREALLOC_BUFFER_MAX	(32 * 1024)


static void fmsi_pcm_free(struct snd_pcm *pcm)
{
	snd_pcm_lib_preallocate_free_for_all(pcm);
}

static int fmsi_pcm_new(struct snd_card *card,
		       struct snd_soc_dai *dai,
		       struct snd_pcm *pcm)
{
	/*
	 * dont use SNDRV_DMA_TYPE_DEV, since it will oops the SH kernel
	 * in MMAP mode (i.e. aplay -M)
	 */
	card->dev->coherent_dma_mask = 0xffffffff;
	return snd_pcm_lib_preallocate_pages_for_all(
		pcm,
		SNDRV_DMA_TYPE_DEV,
		card->dev,
		PREALLOC_BUFFER, PREALLOC_BUFFER_MAX);
}

/************************************************************************


		alsa struct


************************************************************************/
struct snd_soc_dai fmsi_soc_dai = {
		.name			= "FMSI",
		.id			= 0,
		.playback = {
			.rates		= FMSI_RATES,
			.formats	= FMSI_FMTS,
			.channels_min	= 1,
			.channels_max	= 8,
		},
		.capture = {
			.rates		= FMSI_RATES,
			.formats	= FMSI_FMTS,
			.channels_min	= 1,
			.channels_max	= 8,
		},
		.ops = &fmsi_dai_ops,
};
EXPORT_SYMBOL_GPL(fmsi_soc_dai);

struct snd_soc_platform fmsi_soc_platform = {
	.name		= "fmsi-pcm",
	.pcm_ops	= &fmsi_pcm_ops,
	.pcm_new	= fmsi_pcm_new,
	.pcm_free	= fmsi_pcm_free,
};
EXPORT_SYMBOL_GPL(fmsi_soc_platform);

/************************************************************************


		platform function


************************************************************************/
static int fmsi_probe(struct platform_device *pdev)
{
	struct fmsi_master *master;
	int ret;

	dbg_msg_0("fmsi_probe\n");

	if (0 != pdev->id) {
		dev_err(&pdev->dev, "current fmsi support id 0 only now\n");
		return -ENODEV;
	}

	master = kzalloc(sizeof(*master), GFP_KERNEL);
	if (!master) {
		dev_err(&pdev->dev, "Could not allocate master\n");
		ret = -ENOMEM;
		goto exit;
	}

	master->irq[SPU2_USE_DSP_NUM] = platform_get_irq(pdev, 1);
	if ((int)master->irq[SPU2_USE_DSP_NUM] <= 0) {
		dev_err(&pdev->dev, "platform_get_irq() fail.\n");
		ret = -ENODEV;
		goto exit;
	}

	master->info		= pdev->dev.platform_data;
	master->fmsi.base	= master->base;
	master->fmsi.master	= master;

	spin_lock_init(&master->lock);

	pm_runtime_enable(&pdev->dev);
	pm_runtime_resume(&pdev->dev);

	fmsi_soc_dai.dev		= &pdev->dev;
	fmsi_soc_dai.private_data	= &master->fmsi;

	if (!request_mem_region(SPUINIT_SPU_TOP_ADDR,
				SPUINIT_SPU_END_ADDR - SPUINIT_SPU_TOP_ADDR,
				"MFSI-SPU-INIT")) {
		printk(KERN_ERR "request_mem_region error\n");
		ret = -ENOMEM;
		goto exit_kfree;
	}

	spu2_base = ioremap_nocache(SPUINIT_SPU_TOP_ADDR,
				SPUINIT_SPU_END_ADDR - SPUINIT_SPU_TOP_ADDR);
	if (!spu2_base) {
		ret = -ENXIO;
		dev_err(&pdev->dev, "Unable to ioremap SPU2 registers.\n");
		goto exit_kfree;
	}

	ret = request_irq(master->irq[SPU2_USE_DSP_NUM], SPU_INT1,
					IRQF_DISABLED, "spu2_dsp1", master);
	if (ret) {
		dev_err(&pdev->dev, "irq[SPU2_USE_DSP_NUM] request err\n");
		goto exit_iounmap;
	}

	disable_irq(master->irq[SPU2_USE_DSP_NUM]);

	ret = snd_soc_register_platform(&fmsi_soc_platform);
	if (ret < 0) {
		dev_err(&pdev->dev, "cannot snd soc register\n");
		goto exit_free_irq;
	}

	ret = fmsi_spu_sound_mw_init(master);
	if (ret < 0) {
		dev_err(&pdev->dev, "cannot sfmsi_spu_sound_mw_init\n");
		goto exit_free_irq;
	}

	return snd_soc_register_dais(&fmsi_soc_dai, 1);

exit_free_irq:
	free_irq(master->irq[SPU2_USE_DSP_NUM], master);
exit_iounmap:
	iounmap(master->base);
	iounmap(spu2_base);
	pm_runtime_disable(&pdev->dev);
exit_kfree:
	kfree(master);
	master = NULL;
exit:
	return ret;
}

static int fmsi_remove(struct platform_device *pdev)
{
	struct fmsi_master *master;
	int ret;

	master = fmsi_get_master(fmsi_soc_dai.private_data);

	snd_soc_unregister_dais(&fmsi_soc_dai, 1);
	snd_soc_unregister_platform(&fmsi_soc_platform);

	pm_runtime_disable(&pdev->dev);

	free_irq(master->irq[SPU2_USE_DSP_NUM], master);

	iounmap(master->base);
	iounmap(spu2_base);
	kfree(master);

	fmsi_soc_dai.dev		= NULL;
	fmsi_soc_dai.private_data	= NULL;

	ret = fmsi_spu_sound_mw_quit(master);

	return 0;
}

static int fmsi_runtime_nop(struct device *dev)
{
	/* Runtime PM callback shared between ->runtime_suspend()
	 * and ->runtime_resume(). Simply returns success.
	 *
	 * This driver re-initializes all registers after
	 * pm_runtime_get_sync() anyway so there is no need
	 * to save and restore registers here.
	 */
	return 0;
}

static struct dev_pm_ops fmsi_pm_ops = {
	.runtime_suspend	= fmsi_runtime_nop,
	.runtime_resume		= fmsi_runtime_nop,
};

static struct platform_driver fmsi_driver = {
	.driver	= {
		.name	= "sh_fmsi",
		.pm	= &fmsi_pm_ops,
	},
	.probe		= fmsi_probe,
	.remove		= fmsi_remove,
};

static int __init fmsi_mobile_init(void)
{
	return platform_driver_register(&fmsi_driver);
}

static void __exit fmsi_mobile_exit(void)
{
	platform_driver_unregister(&fmsi_driver);
}
module_init(fmsi_mobile_init);
module_exit(fmsi_mobile_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SuperH onchip FMSI audio driver");
MODULE_AUTHOR("Renesas Electronics");
